import numpy as np
from numpy import array, kron
import qutip
from qutip import Qobj, mesolve, lindblad_dissipator, tensor, basis
import numpy.fft as fft
import matplotlib.pyplot as plt

import scipy
from scipy.signal import lfilter
from scipy.linalg import expm
import random

from pathlib import Path

from itertools import product

import h5py
import os


# T1 times in ns
T1ms = np.array([0.833, 0.895, 1.315, 1.496, 0.776, 1.582, 1.439, 0.886, 1.048]) * 1e6


# the cavity modes that we want to use
modelist = [3]



class multimode_system:
    """
    A class to represent a multimode system consisting of a qubit coupled to n modes.
    
    ...
    
    Attributes 
    ----------
    qubit_state_num: int 
        number of qubit levels, i.e. the dimension of the qubit Hilbert space 
        
    mode_state_num: int
        number of levels to simulate for each mode, i.e. maximum photon number at which to truncate
        
    number_of_modes: int 
        the number of modes to model in the system
        
    hparams: dict 
        of the form {"chis": [chi for each mode], "kappas": [kappa for each mode]}, 
        calibration parameters for the system Hamiltonian
        
    coherence_params: dict
        List of relevant coherence parameters. Can contain any of the following keys:
            T1_q: T1 decay time of the qubit, expected type is a float or int
            T2_q: dephasing time of the qubit, expected type is a float or int
            nth_q: thermal population of the qubit, expected type is a float
            T1_ms: T1 decay times of the cavity modes, expected type is a list of floats or ints
            nth_ms: thermal population of the cavity modes, expected type is a list of floats
            
    ROTATING: bool
        True if working in the rotating frame, False otherwise, mostly deprecated and should always
        be True for the Hamiltonian matrix to be accurate
        
    Methods
    -------
    H_rot():
        calculates and returns the bare Hamiltonian, i.e. the Hamiltonian without the drive terms, 
        in the rotating frame and dispersive regime
        
    controlHs():
        calculates and returns the rotation operators that make up the drive terms of the total Hamiltonian
        
    totalH(controls):
        given a set of controls, forms and returns the total Hamiltonian using H_rot() and controlHs()
        
    qutip_mesolve(rho, controls, tlist):
        from a starting state rho, time evolves the system using the Lindblad master equation under a
        set of controls specified at each time point in tlist
    """
    def __init__(self, qubit_state_num, mode_state_num, number_of_modes, hparams, coherence_params=None,
                 ROTATING=True, transfer_fn=False, transfer_fn_files=[]):
        """
        Constructs all the necessary attributes for the multimode_system object and all the relevant
        operators needed to define the Hamiltonian matrix
        """
        self.qnum = qubit_state_num
        self.mnum = mode_state_num
        self.mmnum = number_of_modes
        self.ROTATING = ROTATING
        self.SAMPLE_RATE = 16 # sample rate of the AWG
        self.FIR_MULT = 4
        self.hparams = hparams
        self.coherence_params = coherence_params
        transmon_levels = qubit_state_num
        mode_levels = self.mnum
        # qubit rotation matrix around the x-axis
        self.Q_x = np.diag(np.sqrt(np.arange(1, transmon_levels)), 1) + \
                   np.diag(np.sqrt(np.arange(1, transmon_levels)), -1)
        # qubit rotation matrix around the y-axis
        self.Q_y = (0+1j) * (np.diag(np.sqrt(np.arange(1, transmon_levels)), 1) - \
                             np.diag(np.sqrt(np.arange(1, transmon_levels)), -1))
        # qubit rotation matrix around the z-axis
        self.Q_z = np.diag(np.arange(0, transmon_levels))
        # qubit identity matrix
        self.I_q = np.identity(transmon_levels)
        # cavity mode rotation matrix around the x-axis
        self.M_x = np.diag(np.sqrt(np.arange(1, mode_levels)), 1) + \
                   np.diag(np.sqrt(np.arange(1, mode_levels)), -1)
        # cavity mode rotation matrix around the y-axis
        self.M_y = (0+1j) * (np.diag(np.sqrt(np.arange(1, mode_levels)), 1) - \
                             np.diag(np.sqrt(np.arange(1, mode_levels)), -1))
        # cavity mode rotation matrix around the z-axis
        self.M_z = np.diag(np.arange(0, mode_levels))
        # cavity mode identity matrix
        self.I_m = np.identity(mode_levels)
        # annihilation/lowering operator for a single cavity mode
        self.am =  np.diag(np.sqrt(np.arange(1, mode_levels)), 1)
        # annihilation/lowering operator for the qubit
        self.aq =  np.diag(np.sqrt(np.arange(1, self.qnum)), 1)
        
        # everything defined above was only for each subsystem, to form the real operators that 
        # act on the whole system we need to take tensor products of one of the operators for
        # each of the subsystems in the total system
        
        # for example, if we have 3 modes and want M_x for the second cavity mode, then this is 
        # $I_q \otimes I_1 \otimes M_x \otimes I_3$ 
        
        # arrays of M_z, M_x, M_y, and a for each cavity mode while not acting (having the identity)
        # on each of the other subsystems
        self.M_zs, self.M_xs, self.M_ys, self.ams  =  [],[],[],[]
        # taking the tensor products for the cavity modes
        for k in np.arange(self.mmnum):
            mmz = self.M_z*(k==0) + self.I_m*(1-(k==0))
            mmx = self.M_x*(k==0) + self.I_m*(1-(k==0))
            mmy = self.M_y*(k==0) + self.I_m*(1-(k==0))
            mma = self.am*(k==0) + self.I_m*(1-(k==0))
            for m in np.arange(1, self.mmnum):
                mmz = np.kron(mmz, self.M_z*(k==m) + self.I_m*(1-(k==m)))
                mmx = np.kron(mmx, self.M_x*(k==m) + self.I_m*(1-(k==m)))
                mmy = np.kron(mmy, self.M_y*(k==m) + self.I_m*(1-(k==m)))
                mma = np.kron(mma, self.am*(k==m) + self.I_m*(1-(k==m)))
            self.M_zs.append(mmz)
            self.M_xs.append(mmx)
            self.M_ys.append(mmy)
            self.ams.append(Qobj(np.kron(self.I_q, mma)))
        self.I_mm = self.I_m
        for m in np.arange(1,self.mmnum): 
            self.I_mm = np.kron(self.I_mm, self.I_m)
        # annihilation operator for the qubit tensored with the identity operator for 
        # all the cavity modes
        self.aqmm = Qobj(np.kron(self.aq, self.I_mm))
        self.transfer_fn = transfer_fn
        if transfer_fn:
            self.transfer_fn_files = transfer_fn_files

    
    def H_rot(self):
        """
        Returns the bare Hamiltonian of the system without any of the drive terms.
        This governs the free evolution of the system in the absence of driving.
        """
        chis, kappas = self.hparams["chis"], self.hparams["kappas"]
        if self.ROTATING:
            freq_ge, mode_ens = 0, 0  # GHz, in lab frame
        else:
            freq_ge, mode_ens = 4.9613896, 6.2230641
#             freq_ge, mode_ens = 1.5, 3
            
        chi_mat = np.zeros(self.qnum)
        chi_2_mat = np.zeros(self.qnum)
        try:
            chi_gfs = self.hparams["chi_gfs"]
            alpha = self.hparams['alpha']
        except:
            print("No values for chi_gfs or alpha")
        
        for ii,chi in enumerate(chis):
            chi_mat = np.zeros(self.qnum)
    #         chi_2_mat = np.zeros(self.qnum)

            for ii, chi in enumerate(chis):
                if self.qnum == 2:
                    chi_mat[1] = chi
                elif self.qnum == 3:
                    chi_mat[1] = chi
                    chi_mat[2] = chi_gfs[ii]
                else:
                    raise Exception("Use case of " + str(self.qnum) + " qubit levels not handled yet!")
            # self-Kerr terms
            mode_ens = np.array([2*np.pi*mm*(mode_ens + 0.5*(mm-1)*kappas[ii]) for mm in np.arange(self.mnum)]) 
            # does not include the cross-Kerr interactions between cavity modes
            H_m = np.diag(mode_ens)
            ret = H_m*(ii==0) + self.I_m*(1-(ii==0))
            for m in np.arange(1,self.mmnum):
                ret = np.kron(ret,H_m*(ii==m) + self.I_m*(1-(ii==m)))
            H0 = np.kron(self.I_q, ret)           
            H0 += 2 * 2* np.pi*(np.kron(np.diag(chi_mat), (self.M_zs[ii])))
            if self.qnum == 3:
                H0 += 2 * np.pi * alpha * np.kron(np.diag([0,0,1]), self.I_mm)
        if self.ROTATING:
            return H0
        else:
            q_rot = np.kron(2*np.pi*freq_ge*np.diag(np.arange(self.qnum)), self.I_mm)
            return H0 + q_rot

    
    def controlHs(self):
        """
        Returns the rotation operators that make up the drive terms in our Hamiltonian.
        
        Returns
        -------
        List of rotation operators in the form
        [x qubit rotation, y qubit rotation, mode 1 x rotation, mode 1 y rotation, ..., 
        mode mnum x rotation, mode mnum y rotation]
        
        """
        controlHs = []
        XI = np.kron(self.Q_x, self.I_mm)
        YI = np.kron(self.Q_y, self.I_mm)
        controlHs.append(XI)
        if self.ROTATING:
            controlHs.append(YI)
        for m in np.arange(self.mmnum):
            IX = np.kron(self.I_q, self.M_xs[m])
            IY = np.kron(self.I_q, self.M_ys[m])
            controlHs.append(IX)
            if self.ROTATING:
                controlHs.append(IY)
        return controlHs
    
    
    def cubic_interpolate_16(self, controls, tlist):
        '''Cubic interpolate the pulse at 1/16 ns resolution'''
        cubic_pulses_16 = []
        total_time = tlist[-1]
        steps = int(np.floor(self.SAMPLE_RATE*total_time))
        tlist_16 = np.array([i/self.SAMPLE_RATE for i in range(steps+1)])
        for base_pulse in controls:
            cubic_spline = scipy.interpolate.splrep(tlist, base_pulse)
            cubic_pulse = scipy.interpolate.splev(tlist_16, cubic_spline, der = 0)
            cubic_pulses_16.append(cubic_pulse)
        return cubic_pulses_16, tlist_16
            
    def interpolate_controls(self, controls, tlist):
        exfine_pulses = [] #finer resolution pulses for qutip mesolver
        num_ops = len(controls) #number of operators (pulses)
        total_time = tlist[-1]
        steps = len(tlist)
        dt = float(total_time) / steps
        fine_steps = int(np.ceil(self.SAMPLE_RATE*total_time))
        exfine_steps = fine_steps * 16
        base_times = np.linspace(0.0, total_time, steps) 
        fine_tlist = np.linspace(0., total_time, fine_steps)
        exfine_tlist = np.linspace(0., total_time, exfine_steps)
        for i in range(num_ops):
            base_pulse = controls[i]  # final control pulse
            #base_pulse = np.append(base_pulse, 0.0)  # add extra 0 on end of pulses for interpolation
            interpFun = scipy.interpolate.interp1d(base_times, base_pulse, kind = "cubic")
            pulse_interp = interpFun(fine_tlist)
            interFun_ef = scipy.interpolate.interp1d(fine_tlist, pulse_interp, kind = "zero")
            exfine_pulse_interp = interFun_ef(exfine_tlist)
            exfine_pulses.append(exfine_pulse_interp) #nopsxfine_steps matrix containing all the pulses
        return exfine_pulses, exfine_tlist
    
    def interpolate_controls_fir(self, controls, tlist):
        '''FIR filter after pulses have already been cubic interpolated at 1/16 ns'''
        fir_coeffs = [-0.00458,-0.00529,-0.00437,0,0.00625,0.01054,0.00884,0,-0.01222,-0.02015,
                      -0.01654,0,0.02201,0.03572,0.02891,0,-0.03769,-0.06077,-0.049,0,0.06403,0.10402,
                      0.08498,0,-0.11659,-0.19748,-0.1713,0,0.29478,0.63158,0.89853,1,0.89853,0.63158,0.29478,
                      0,-0.1713,-0.19748,-0.11659,0,0.08498,0.10402,0.06403,0,-0.049,-0.06077,-0.03769,0,
                      0.02891,0.03572,0.02201,0,-0.01654,-0.02015,-0.01222,0,0.00884,0.01054,0.00625,0,
                      -0.00437,-0.00529,-0.00458,0]
        fir_pulses = []
        steps = (len(tlist)-1)*self.FIR_MULT
        t_indices = [i*self.FIR_MULT for i in range(len(tlist))]
        tlist_64 = np.array([i/(self.SAMPLE_RATE*self.FIR_MULT) for i in range(steps + 1)])

        for pulse_16 in controls: 
            awg_pulse = np.zeros(len(tlist_64))
            for j in range(len(pulse_16)):
                awg_pulse[t_indices[j]] = pulse_16[j]
            fir_pulse = lfilter(fir_coeffs, [1] + steps*[0.], awg_pulse)  # maybe should be fir_coeffs[::-1]
            fir_pulses.append(fir_pulse)
        return fir_pulses, tlist_64
    
#         exfine_pulses = [] #finer resolution pulses for qutip mesolver
#         num_ops = len(controls) #number of operators (pulses)
#         total_time = tlist[-1]
#         steps = len(tlist)
#         dt = float(total_time) / steps
#         fine_steps = int(np.ceil(self.SAMPLE_RATE*total_time))
#         exfine_steps = fine_steps * 16
#         base_times = np.linspace(0.0, total_time, steps) 
#         fine_tlist = np.linspace(0., total_time, fine_steps)
#         exfine_tlist = np.linspace(0., total_time, exfine_steps)
#         for i in range(num_ops):
#             base_pulse = controls[i]  # final control pulse
#             #base_pulse = np.append(base_pulse, 0.0)  # add extra 0 on end of pulses for interpolation
#             interpFun = scipy.interpolate.interp1d(base_times, base_pulse, kind = "cubic")
#             pulse_interp = interpFun(fine_tlist)
        
        
#         num_ops = len(controls) #number of operators (pulses)
#         total_time = tlist[-1]
#         t_indices = [int(np.floor(t*self.SAMPLE_RATE)) for t in tlist]
#         steps = int(np.floor(self.SAMPLE_RATE*total_time))
#         exfine_steps = steps*16
#         fine_tlist = np.array([i/self.SAMPLE_RATE for i in range(steps+1)])
#         exfine_tlist = np.linspace(0., fine_tlist[-1], exfine_steps)
#         exfine_pulses = []
#         print("FIR Interpolation")
#         for i in range(num_ops):
#             base_pulse = controls[i]
#             awg_pulse = np.zeros(steps + 1)
#             for j in range(len(base_pulse)): 
#                 awg_pulse[t_indices[j]] = base_pulse[j]
            
#             fir_pulse = lfilter(fir_coeffs, [1] + steps*[0.], awg_pulse)
#             print(fir_pulse[0:50])   
#             interFun_ef = scipy.interpolate.interp1d(fine_tlist, fir_pulse, kind = "zero")
#             exfine_pulse_interp = interFun_ef(exfine_tlist)
#             exfine_pulses.append(fir_pulse)
                
#         return exfine_pulses, fine_tlist

    def quantize_controls(self, controls, qubit_resolution=0.0011811023622, cavity_resolution=0.00151968503937, scheme='abs'):
        """
        Function that simulates the quantization effects of the control electronics controls
        used in the experiment.

        controls: List
            a list of the form [[qubit control x], [qctrl y], [mode1cntrlx], [mode1cntrly], ...] 
            where the lists are the values of the controls at each time step

        resolution: float
            The resolution or quantization step size of the control electronics

        """
        
        scheme_types = ['abs', 'floor', 'nearest_int']
        if scheme not in scheme_types:
            raise ValueError("Invalid quantization scheme type. Expected one of: %s" % scheme_types)
        
        quantized_controls = []
        if scheme == 'abs':
            for i, control_array in enumerate(controls):
                signs = np.sign(control_array)
                if i == 0 or i == 1:
                    quantized_array = [np.abs(control_point) // qubit_resolution * qubit_resolution 
                                       for control_point in control_array]
                    quantized_array = np.array(quantized_array) * signs
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.abs(control_point) // cavity_resolution * cavity_resolution 
                                       for control_point in control_array]
                    quantized_array = np.array(quantized_array) * signs
                    quantized_controls.append(quantized_array)
            return np.array(quantized_controls)
        elif scheme == 'nearest_int': 
            for i, control_array in enumerate(controls):
                if i == 0 or i == 1:
                    quantized_array = [np.floor(control_point / qubit_resolution + 1/2) * qubit_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.floor(control_point / cavity_resolution + 1/2) * cavity_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
            return np.array(quantized_controls)
        else:
            for i, control_array in enumerate(controls):
                if i == 0 or i == 1:
                    quantized_array = [np.floor(control_point / qubit_resolution) * qubit_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.floor(control_point / cavity_resolution) * cavity_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
            return np.array(quantized_controls)
        
    def implement_transfer_fn(self, control, filename, freq_to_amp=True):
        # add current directory to path
        dir = "/home/aaron/projects/IterativeLearningControl.jl/examples/notebooks/multimode"
        file = os.path.join(dir, filename)
        with h5py.File(file, 'r') as f:
            if freq_to_amp:
                omegas = f['omegas'][()]
                amps = f['amps'][()]
            else:  # inverse transfer function
                omegas = f['amps'][()]
                amps = f['omegas'][()]
        # assume zero frequency at zero amplitude, used for interpolation function
        omegas = np.append(omegas, -omegas)
        amps = np.append(amps, -amps)
        omegas = np.append(omegas, 0.0)
        amps = np.append(amps, 0.0)
        o_s = [x for y, x in sorted(zip(amps, omegas))]
        a_s = np.sort(amps)

        # interpolate data, transfer_fn is a function that for each omega returns the corresponding amp
        transfer_fn = scipy.interpolate.interp1d(o_s, a_s)
        output_amps = []
        max_interp_index = np.argmax(omegas)
        for i in range(len(control)):
            # if frequency greater than calibrated range, assume a proportional relationship (high amp)
            if np.abs(control[i]) > omegas[max_interp_index]:
                output_amps.append(control[i] * amps[max_interp_index] / omegas[max_interp_index])
                # print("!Interpolating beyond max amp!")

            else:  # otherwise just use the interpolated transfer function
                output_amps.append(transfer_fn((control[i])))
        return np.array(output_amps)
    

    def plot_pulses(self, controls, tlist, plot_qubit=True, interp=False):
        if interp:
            pulses, tlist = self.interpolate_controls(controls, tlist)
        else:
            pulses = controls
        fig, ax = plt.subplots(nrows=1, figsize=(14,4))
        if plot_qubit: 
            start = 0
        else:
            start = int(2)
        if self.ROTATING:
            labels = ["Qubit_x", "Qubit_y", "Cavity_x", "Cavity_y"]
        else:
            labels = ["Qubit", "Cavity"]
        for ii, x in enumerate(pulses[start:]):
            ax.plot(tlist/1e3, (x/2/np.pi)*1e3, label=labels[ii+start]) # plotting the linear frequency in MHz
        ax.set_xlabel("Time ($\\mu$s)", fontsize=20)
        ax.set_ylabel("Drive Strength (MHz)", fontsize=20)
        ax.tick_params(direction='in', length=6, width=2, colors='k', \
                grid_color='r', grid_alpha=0.5, labelsize=14, labelbottom=True, right=True, top=True)
        ax.legend(loc='best', fontsize=12)
        
    def total_H(self, controls):
        """
        Returns the total Hamiltonian at each time step in a format that can be fed into qutip.mesolve().
        This defines a time-dependent Hamiltonians in QuTiP using the method in [0].
        
        [0] https://qutip.org/docs/latest/guide/dynamics/dynamics-time.html#:~:text=Array%20Based%3A%20The,the%20given%20times.
        
        Parameters 
        ----------
        controls: List
            a list of the form [[qubit control x], [qctrl y], [mode1cntrlx], [mode1cntrly], ...] where the lists
            are floats that model the values of the controls at each time step
    
        Returns
        -------
        A list of the form 
        [H_bare, [H_drive_qubit_x, [control coefficient for qubit rotation about x at each time step], 
        [H_drive_qubit_y, [control coeffs at each time step]], ...]
        
        """
        H = [Qobj(self.H_rot())]
        controlHs = self.controlHs()
        #controls give the coefficient value for each of the controlHs at each time step
        
        for index in range(len(controls)):
            H.append([Qobj(controlHs[index]), controls[index]])
        return H
    
    def qutip_mesolve(self, rho, controls, tlist, state=False, quantize=True, fir=True, driver_scheme='abs', awg_scheme='nearest_int', plot=False, verbose=False):
        """
        From a starting density matrix rho, time evolves the system using the Lindblad master equation under a
        set of controls specified at each time point in tlist
        
        Parameters
        ----------
        rho: QuTiP Qobj
            input state density matrix or state vector if state=true
    
        controls: List 
            of the form [[qubit control x], [qctrl y], [mode1cntrlx], [mode1cntrly], ...] where the lists
            are the values of the controls at each time in tlist
            
        tlist: List
            list of times (dtype float) starting from 0.0 to the final time, such as in increments of some
            dt, although this is not strictly required.
        
        Returns
        -------
        A QuTiP Result object from which we can get the density matrix at the final time in tlist.
        
        """
        
        freq_ge, mode_freq = 4.9613896, 6.2230641
#         freq_ge, mode_freq = 1.5, 3
        
        if state and not np.shape(rho)[1]==1:
            raise Exception("State is true but initial state is not a vector")
        scale_factor = 2
        if self.transfer_fn:
            if verbose:
                print("Implementing Transfer Function")
            transfer_fn_controls = []
            for j in range(len(controls)):
                transfer_fn_controls.append(self.implement_transfer_fn(controls[j], self.transfer_fn_files[j//2],
                                                                      freq_to_amp=True))
            controls = transfer_fn_controls
            scale_factor = 1
                   
        
        if fir: 
            #cubic interpolate the amplitudes           
            controls, tlist = self.cubic_interpolate_16(controls, tlist)
            #modulate
            if not self.ROTATING:
                actual_controls = []

                actual_controls.append(scale_factor*controls[0]*np.cos(2*np.pi*freq_ge*tlist)
                                       -scale_factor*controls[1]*np.sin(2*np.pi*freq_ge*tlist))

                actual_controls.append(scale_factor*controls[2]*np.cos(2*np.pi*mode_freq*tlist) 
                                       -scale_factor*controls[3]*np.sin(2*np.pi*mode_freq*tlist))
                controls = actual_controls
            #Quantize the modulation to feed into AWG
            if verbose:
                print("Quantizing Amplitudes")
            controls = self.quantize_controls(controls, qubit_resolution=1/127, cavity_resolution=1/127, scheme=driver_scheme)
            
            #FIR filter
            if verbose:
                print("Applying FIR Filter") 
            controls, tlist = self.interpolate_controls_fir(controls, tlist)
        
        else: 
            if not self.ROTATING:
                actual_controls = []

                actual_controls.append(scale_factor*controls[0]*np.cos(2*np.pi*freq_ge*tlist)
                                       -scale_factor*controls[1]*np.sin(2*np.pi*freq_ge*tlist))

                actual_controls.append(scale_factor*controls[2]*np.cos(2*np.pi*mode_freq*tlist) 
                                       -scale_factor*controls[3]*np.sin(2*np.pi*mode_freq*tlist))
                controls = actual_controls
                
                controls, tlist = self.interpolate_controls(controls, tlist)
            
        if quantize:
            if verbose:
                print("Quantizing Controls after FIR Filter")
            if not self.transfer_fn:
                controls = self.quantize_controls(controls, scheme=awg_scheme)
            else:
                controls = self.quantize_controls(controls, qubit_resolution=1/127, cavity_resolution=1/127, scheme=awg_scheme)
        
        if self.transfer_fn and plot:
            fig = plt.figure(figsize=(14,5))
            ax = fig.add_subplot(111)
            for j in range(len(controls)):
                ax.plot(tlist, controls[j])
            ax.set_xlabel("Time (ns)")
            ax.set_ylabel("Amp")
            
        if self.transfer_fn:
            transfer_fn_controls = []
            for j in range(len(controls)):
                transfer_fn_controls.append(self.implement_transfer_fn(controls[j], self.transfer_fn_files[j//2],
                                                                      freq_to_amp=False))
            controls = 2 * np.array(transfer_fn_controls)
                                                
        if plot:
            self.plot_pulses(controls, tlist, interp=False)
        H = self.total_H(controls)
        #print(H)
        if state:
            out = qutip.sesolve(H, Qobj(rho), tlist)
        else:
            if self.coherence_params is None:
                c_ops = []
            else:
                gamma = 1/self.coherence_params['T1_q']
                gamma_phi = 1/self.coherence_params['T2_q'] - 1/2/self.coherence_params['T1_q'] 
                n_thq = self.coherence_params['nth_q']
                c_ops = gamma*(1+n_thq)*lindblad_dissipator(self.aqmm) + \
                        gamma * (n_thq) * lindblad_dissipator(self.aqmm.dag())
                c_ops += gamma_phi * lindblad_dissipator(self.aqmm.dag() * self.aqmm) 
                kappa_ms = 1/array(self.coherence_params['T1_ms'])
                kappa_phi_ms = 1/array(self.coherence_params['T2_ms']) - 1/2/array(self.coherence_params['T1_ms'])
                n_thms =  array(self.coherence_params['nth_ms'])
                for ii,a in enumerate(self.ams):
                    c_ops += kappa_ms[ii]*(1 + n_thms[ii]) * lindblad_dissipator(a) + \
                             kappa_ms[ii]*(n_thms[ii]) * lindblad_dissipator(a.dag()) 
                    c_ops += kappa_phi_ms[ii]*lindblad_dissipator(a.dag()*a)
            out = mesolve(H, Qobj(rho), tlist, c_ops=c_ops) #options=qutip.Options(nsteps=10000)
        return out   
    

# def quantize_controls(controls, qubit_resolution=0.00088357293, cavity_resolution=0.000234375):
#     """
#     Function that simulates the quantization effects of the control electronics controls
#     used in the experiment.
    
#     controls: List
#         a list of the form [[qubit control x], [qctrl y], [mode1cntrlx], [mode1cntrly], ...] 
#         where the lists are the values of the controls at each time step
        
#     resolution: float
#         The resolution or quantization step size of the control electronics
    
#     """
#     quantized_controls = []
#     for i, control_array in enumerate(controls):
#         if i == 0 or i == 1:
#             quantized_array = [np.floor(control_point // qubit_resolution) * qubit_resolution for control_point in control_array]
#             quantized_controls.append(quantized_array)
#         else:
#             quantized_array = [np.floor(control_point // cavity_resolution) * cavity_resolution for control_point in control_array]
#             quantized_controls.append(quantized_array)
#     return np.array(quantized_controls)

# def quantize_controls(controls, qubit_resolution=0.0011811023622, cavity_resolution=0.00151968503937, scheme='abs'):
#     """
#     Function that simulates the quantization effects of the control electronics controls
#     used in the experiment.

#     controls: List
#         a list of the form [[qubit control x], [qctrl y], [mode1cntrlx], [mode1cntrly], ...] 
#         where the lists are the values of the controls at each time step

#     resolution: float
#         The resolution or quantization step size of the control electronics

#     """
    
#     scheme_types = ['abs', 'floor', 'nearest_int']
#     if scheme not in scheme_types:
#         raise ValueError("Invalid quantization scheme type. Expected one of: %s" % scheme_types)
    
#     quantized_controls = []
#     if scheme == 'abs':
#         for i, control_array in enumerate(controls):
#             signs = np.sign(control_array)
#             if i == 0 or i == 1:
#                 quantized_array = [np.abs(control_point) // qubit_resolution * qubit_resolution 
#                                     for control_point in control_array]
#                 quantized_array = np.array(quantized_array) * signs
#                 quantized_controls.append(quantized_array)
#             else:
#                 quantized_array = [np.abs(control_point) // cavity_resolution * cavity_resolution 
#                                     for control_point in control_array]
#                 quantized_array = np.array(quantized_array) * signs
#                 quantized_controls.append(quantized_array)
#         return np.array(quantized_controls)
#     elif scheme == 'nearest_int': 
#         for i, control_array in enumerate(controls):
#             if i == 0 or i == 1:
#                 quantized_array = [np.floor(control_point / qubit_resolution + 1/2) * qubit_resolution for control_point in control_array]
#                 quantized_controls.append(quantized_array)
#             else:
#                 quantized_array = [np.floor(control_point / cavity_resolution + 1/2) * cavity_resolution for control_point in control_array]
#                 quantized_controls.append(quantized_array)
#         return np.array(quantized_controls)
#     else:
#         for i, control_array in enumerate(controls):
#             if i == 0 or i == 1:
#                 quantized_array = [np.floor(control_point / qubit_resolution) * qubit_resolution for control_point in control_array]
#                 quantized_controls.append(quantized_array)
#             else:
#                 quantized_array = [np.floor(control_point / cavity_resolution) * cavity_resolution for control_point in control_array]
#                 quantized_controls.append(quantized_array)
#         return np.array(quantized_controls)

def quantize_controls(controls, qubit_resolution=0.0011811023622, cavity_resolution=0.00151968503937, scheme='abs', rotating=False):
    """
    Function that simulates the quantization effects of the control electronics controls
    used in the experiment.

    controls: List
        a list of the form [[qubit control x], [qctrl y], [mode1cntrlx], [mode1cntrly], ...] 
        where the lists are the values of the controls at each time step

    resolution: float
        The resolution or quantization step size of the control electronics

    """
    
    scheme_types = ['abs', 'floor', 'nearest_int']
    if scheme not in scheme_types:
        raise ValueError("Invalid quantization scheme type. Expected one of: %s" % scheme_types)
    
    quantized_controls = []
    if scheme == 'abs':
        for i, control_array in enumerate(controls):
            signs = np.sign(control_array)
            if rotating:
                if i == 0 or i == 1:
                    quantized_array = [np.abs(control_point) // qubit_resolution * qubit_resolution 
                                        for control_point in control_array]
                    quantized_array = np.array(quantized_array) * signs
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.abs(control_point) // cavity_resolution * cavity_resolution 
                                        for control_point in control_array]
                    quantized_array = np.array(quantized_array) * signs
                    quantized_controls.append(quantized_array)
            else: 
                if i == 0:
                    quantized_array = [np.abs(control_point) // qubit_resolution * qubit_resolution 
                                        for control_point in control_array]
                    quantized_array = np.array(quantized_array) * signs
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.abs(control_point) // cavity_resolution * cavity_resolution 
                                        for control_point in control_array]
                    quantized_array = np.array(quantized_array) * signs
                    quantized_controls.append(quantized_array)
        return np.array(quantized_controls)
    elif scheme == 'nearest_int': 
        for i, control_array in enumerate(controls):
            if rotating:
                if i == 0 or i == 1:
                    quantized_array = [np.floor(control_point / qubit_resolution + 1/2) * qubit_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.floor(control_point / cavity_resolution + 1/2) * cavity_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
            else:
                if i == 0:
                    quantized_array = [np.floor(control_point / qubit_resolution + 1/2) * qubit_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.floor(control_point / cavity_resolution + 1/2) * cavity_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
        return np.array(quantized_controls)
    else:
        for i, control_array in enumerate(controls):
            if rotating:
                if i == 0 or i == 1:
                    quantized_array = [np.floor(control_point / qubit_resolution) * qubit_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.floor(control_point / cavity_resolution) * cavity_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
            else: 
                if i == 0:
                    quantized_array = [np.floor(control_point / qubit_resolution) * qubit_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
                else:
                    quantized_array = [np.floor(control_point / cavity_resolution) * cavity_resolution for control_point in control_array]
                    quantized_controls.append(quantized_array)
        return np.array(quantized_controls)


def plot_pops(tlist, states, qubit_levels, mode_levels):
    fig, ax = plt.subplots(figsize=(14,8))
    pops = []
    for j in range(mode_levels):
        for i in range(qubit_levels):
            psi = tensor(basis(qubit_levels, i), basis(mode_levels,j))
            proj = psi * psi.dag()
            pop_ij = [np.real(np.trace(np.dot(proj, state))) for state in states]
            pops.append(pop_ij)
    labels = []
    for j in range(mode_levels):
        labels.append("g{}".format(j))
        labels.append("e{}".format(j))
        labels.append("f{}".format(j))
    for ii, pop in enumerate(pops):
        ax.plot(tlist, pop, label=labels[ii])
    ax.set_xlabel("Time (ns)")
    ax.set_ylabel("Population")
#     ax.tick_params(direction='in', length=6, width=2, colors='k', \
#             grid_color='r', grid_alpha=0.5, labelsize=14, labelbottom=True, right=True, top=True)
    ax.legend(loc='best')

def time_evol_f(
    rho, 
    controls, 
    tlist, 
    # taus, 
    qubit_state_num, 
    mode_state_num, 
    number_of_modes, 
    hparams, 
    state=False, 
    coherence_params=None, 
    rotating=True, 
    plot=False, 
    quantize=True, 
    fir=True,
    transfer_fn=False, 
    transfer_fn_files=[], 
    driver_scheme='abs', 
    awg_scheme='nearest_int', 
    verbose=False
):
    """
    Function that, given an input density matrix (state) rho, evolves rho under controls, returning the
    final density matrix at the last time in tlist
    
    rho: input state matrix, as a 2D numpy array or state vector if state=True
    
    controls: a list of the form [[qubit control x], [qctrl y], [mode1cntrlx], [mode1cntrly], ...] where the lists
    are the values of the controls at each time in tlist
    
    tlist: List
        list of times (dtype float) starting from 0.0 to the final time, such as in increments of some
        dt although this is not strictly required.
        
    qubit_state_num: int 
        number of qubit levels i.e. the dimension of the qubit Hilbert space 
        
    mode_state_num: int
        number of levels to simulate for each mode 
        
    number_of_modes: int 
        the number of modes the qubit is coupled to 
        
    hparams: dict 
        of the form {"chis": [chi for each mode], "kappas": [kappa for each mode]}, 
        calibration parameters for the Hamiltonian
        
    coherence_params: dict
        List of relevant coherence parameters. Can contain any of the following keys:
            T1_q: T1 decay time of the qubit
            T2_q: dephasing time of the qubit
            nth_q: thermal population of the qubit
            T1_ms: T1 decay times of the cavity modes, expected type is a list
            nth_ms: thermal population of the cavity modes, expected type is a list
    """
    if type(rho) == qutip.qobj.Qobj:
        rho = rho.full()  # convert rho to a numpy array if it is specified as a Qobj
        
    msys = multimode_system(qubit_state_num, mode_state_num, number_of_modes, hparams, 
                            coherence_params=coherence_params, ROTATING=rotating, transfer_fn=transfer_fn,
                           transfer_fn_files=transfer_fn_files)
    #msys.plot_pulses(controls, tlist)
    out = msys.qutip_mesolve(rho, controls, tlist, state=state, quantize=quantize, fir=fir, driver_scheme=driver_scheme, awg_scheme=awg_scheme, plot=plot, verbose=verbose)
    # return the final state density matrix at the final time in tlist
    if plot:
        plot_pops(out.times, out.states, qubit_state_num, mode_state_num)
    return out.states[-1]

# time evolve the intial state under the controls
hparams = {
    "chis": [-0.5469e-3], 
    "kappas": [4e-6], 
    "chi_gfs": [-1.015e-3], 
    "alpha": -0.143
}

coherence_params = {
    "T1_q": 100e3, 
    "T2_q": 150e3, 
    "nth_q": 0.015, 
    "T1_ms": T1ms[modelist], 
    "nth_ms": [0.005], 
    "T2_ms": [2e6]
}

transfer_fn_files_niq = [
    "00000_strong_qubit_transfer_calibration_NIQ.h5", 
    "00000_cavity_strong_transfer_calibration_mode3_NIQ.h5"
]

transfer_fn_files_abs = [
    "00003_strong_qubit_transfer_calibration.h5",
    "00003_cavity_strong_transfer_calibration_mode3.h5"
]



# prefix files with directory of file
# transfer_fn_files = [os.path.join(os.path.dirname(os.path.abspath(__file__)), file) for file in transfer_fn_files]

mode_levels = 14
qubit_levels = 3

# initial state |g0>
psi = tensor(basis(qubit_levels, 0), basis(mode_levels, 0))

def lab_frame_experiment(controls, times, taus, scheme='nearest_int', verbose=False):

    if scheme == 'nearest_int':
        transfer_fn_files = transfer_fn_files_niq
    elif scheme == 'abs':
        transfer_fn_files = transfer_fn_files_abs
    else:
        raise ValueError("Invalid quantization scheme")

    # tau_times = [times[tau] for tau in taus]

    out_l = time_evol_f(
        psi, 
        controls, 
        times, 
        # tau_times,
        qubit_levels, 
        mode_levels, 
        len(modelist), 
        hparams, 
        coherence_params=coherence_params, 
        state=True,
        rotating=False, 
        plot=False, 
        quantize=True, 
        fir=True,
        transfer_fn=True, 
        transfer_fn_files=transfer_fn_files, 
        driver_scheme=scheme,
        verbose=verbose
    )

    return out_l

