using QuantumCollocation
using NamedTrajectories
using IterativeLearningControl
const ILC = IterativeLearningControl

Z = load_traj(joinpath(@__DIR__, "../notebooks/multimode/trajectories/g0_to_g1.jld2"))

linsys_solver = "qdldl"

transmon_levels = 3
cavity_levels = 14

system = MultiModeSystem(
    transmon_levels,
    cavity_levels
)

P = FourthOrderPade(system)


function f(
    zₜ::AbstractVector{<:Real},
    zₜ₊₁::AbstractVector{<:Real}
)
    ψ̃ₜ = zₜ[Z.components.ψ̃]
    ψ̃ₜ₊₁ = zₜ₊₁[Z.components.ψ̃]
    aₜ = zₜ[Z.components.a]
    aₜ₊₁ = zₜ₊₁[Z.components.a]
    daₜ = zₜ[Z.components.da]
    daₜ₊₁ = zₜ₊₁[Z.components.da]
    ddaₜ = zₜ[Z.components.dda]
    Δt = Z.timestep
    Gₜ = QuantumCollocation.Integrators.G(aₜ, system.G_drift, system.G_drives)
    return [
        ψ̃ₜ₊₁ - fourth_order_pade(Gₜ * Δt) * ψ̃ₜ;
        aₜ₊₁ - aₜ - Δt * daₜ;
        daₜ₊₁ - daₜ - Δt * ddaₜ;
    ]
end


g(ψ̃) = abs2.(iso_to_ket(ψ̃))

a_bounds = Z.bounds.a[2]

function quantize(A::Matrix{Float64})
    res = a_bounds ./ 128
    Â = zeros(size(A))
    for i = axes(A, 2)
        a_quantized = round.(A[:, i] ./ res) .* res
        Â[:, i] = a_quantized
    end
    return Â
end

# τs = 1:Ẑ.T
τs = 1:Z.T

experiment = QuantumSimulationExperiment(
    system,
    Z[1].ψ̃,
    g,
    τs;
    integrator=exp,
    drive_transform=quantize,
)

max_iter = 5
max_backtrack_iter = 15
fps = 2
α = 0.5
β = 1.0
R = (
    a   = 1.0e1,
    dda = 1.0e10
)
Qy = 1.0e-5
Qyf = 1.0e3

QP_tol = 1e-6
QP_max_iter = 1e5
QP_settings = Dict(
    :polish => true,
    :linsys_solver => linsys_solver,
)

static = false
nominal_correction = false
dynamic_measurements = false
correction_term = true

prob = ILCProblem(
    Z,
    f,
    experiment;
    max_iter=max_iter,
    QP_verbose=false,
    correction_term=correction_term,
    dynamic_measurements=dynamic_measurements,
    norm_p=1,
    R=R,
    static_QP=static,
    Qy=Qy,
    Qyf=Qyf,
    α=α,
    β=β,
    max_backtrack_iter=max_backtrack_iter,
    QP_tol=QP_tol,
    QP_max_iter=QP_max_iter,
    QP_settings=QP_settings,
    mle=true,
)

ILC.solve!(prob; nominal_correction=nominal_correction)

plot_dir = "plots/multimode/ILC/refactor"

plot_name =
    (static ? "static_" : "dynamic_") *
    "T_$(Z.T)_taus_$(length(τs))_" *
    "alpha_$(α)_beta_$(β)_" *
    "R_" * join([string(k) * "_" * string(v) for (k, v) in pairs(R)], "_") * "_" *
    "Qy_$(Qy)_Qyf_$(Qyf)_QP_tol_$(QP_tol)" *
    (nominal_correction ? "_nominal_correction" : "")

plot_path = generate_file_path("gif", plot_name, plot_dir)

animate_ILC_multimode_talk(prob, plot_path; fps=fps, plot_transformed_controls=true)
