using CairoMakie
using LinearAlgebra

using PicoOld
using NamedTrajectories
using IterativeLearningControl
const ILC = IterativeLearningControl

# get path above current directory
traj = load_traj(joinpath(@__DIR__, "../notebooks/multimode/trajectories/g0_to_g1.jld2"))


ketdim = traj.dims.ψ̃ ÷ 2

function pops(ψ̃)
    ψ = iso_to_ket(ψ̃)
    return abs2.(ψ)
end

# τs = [traj.T]
τs = 1:traj.T

transmon_levels = 3
cavity_levels = 14

dynamics_system = MultiModeSystem(transmon_levels, cavity_levels, "g0", "g1")
system = MultiModeSystem(transmon_levels, cavity_levels, "g0", "g1")

P = FourthOrderPade(system)

function dynamics(
    zₜ::AbstractVector{<:Real},
    zₜ₊₁::AbstractVector{<:Real},
    Z::NamedTrajectory
)
    xₜ = zₜ[Z.components.states]
    xₜ₊₁ = zₜ₊₁[Z.components.states]
    uₜ = zₜ[Z.components.controls]
    Δt = Z.timestep
    return QuantumCollocation.Dynamics.dynamics(xₜ₊₁, xₜ, uₜ, Δt, P, dynamics_system)
end

a_bounds = traj.bounds.a

function quantize(A::Matrix{Float64}; scheme=floor)
    res = a_bounds ./ 128
    Â = similar(A)
    for i = 1:size(A, 2)
        Â[:, i] = scheme.(A[:, i] ./ res) .* res
    end
    return Â
end

experiment = QuantumSimulationExperiment(
    system,
    traj[1].ψ̃,
    pops,
    τs;
    integrator=exp,
    drive_transform=quantize,
)

max_iter = 20
max_backtrack_iter = 20
fps = 2
α = 0.5
β = 0.5^6
R = (
    a   = 1.0e1,
    dda = 1.0e10
)
Qy = 1.0e-5
Qyf = 1.0e3
QP_tol = 1e-6
QP_max_iter = 1e5
QP_linear_solver = "mkl pardiso"

QP_settings = Dict(
    :polish => true,
)

static = false
nominal_correction = false
dynamic_measurements = true
correction_term = true

norm_p = 1

y_final = [0; 1; zeros(ketdim - 2)]

prob = ILCProblem(
    traj,
    dynamics,
    experiment;
    max_iter=max_iter,
    QP_verbose=false,
    correction_term=correction_term,
    dynamic_measurements=dynamic_measurements,
    norm_p=norm_p,
    R=R,
    static_QP=static,
    Qy=Qy,
    Qyf=Qyf,
    α=α,
    β=β,
    max_backtrack_iter=max_backtrack_iter,
    QP_tol=QP_tol,
    QP_max_iter=QP_max_iter,
    QP_settings=QP_settings,
    QP_linear_solver=QP_linear_solver,
    mle=true,
    # y_final=y_final,
)

ILC.solve!(prob;
    nominal_correction=nominal_correction,
    show_metric=true,
    # metric=:full_state,
)

plot_dir = joinpath(@__DIR__, "plots")

plot_name =
    (static ? "static_" : "dynamic_") *
    "T_$(traj.T)_taus_$(length(τs))_" *
    "alpha_$(α)_beta_$(β)_" *
    "R_" * join([string(k) * "_" * string(v) for (k, v) in pairs(R)], "_") * "_" *
    "Qy_$(Qy)_Qyf_$(Qyf)_QP_tol_$(QP_tol)" *
    (nominal_correction ? "_nominal_correction" : "")

plot_path = generate_file_path("gif", plot_name, plot_dir)

animate_ILC_multimode_talk(prob, plot_path; fps=fps, plot_transformed_controls=true)
