@info "Threads available: $(Threads.nthreads())"

using ExponentialAction
using PyCall
using BenchmarkTools

using NamedTrajectories
using QuantumCollocation
using IterativeLearningControl
const ILC = IterativeLearningControl

@pyinclude joinpath(@__DIR__, "lab_frame_quantization.py")

function quantize_controls(Ã::AbstractMatrix)
    as = py"quantize_controls"(Ã)
    return as
end

traj = load_traj(joinpath(@__DIR__, "../../data/multimode/g0_to_g1.jld2"))

rotating_frame_system = MultiModeSystem(3, 14)

rotating_frame_experiment = QuantumSimulationExperiment(
    rotating_frame_system,
    traj[1].ψ̃,
    ψ̃ -> ψ̃,
    [length(times(traj))];
    integrator=exp,
)

Ψ = nominal_rollout(traj.a, times(traj), rotating_frame_experiment; integrator=exp)

populations(Ψ[:, end])


ct = ContinuousTrajectory(traj.a, traj.da, times(traj))


timestep = 1/2^6

# timestep = 1/16

ts = 0:timestep:times(traj)[end]

A = ct(ts)

lab_frame_system = MultiModeSystem(3, 14; lab_frame=true)

ωq = lab_frame_system.params[:ωq]
ωc = lab_frame_system.params[:ωc]

function to_lab_frame(a::AbstractVector, t::Real)
    a_q_x = a[1]
    a_q_y = a[2]

    a_c_x = a[3]
    a_c_y = a[4]

    a_q = 2 * a_q_x * cos(ωq * t) - 2 * a_q_y * sin(ωq * t)
    a_c = 2 * a_c_x * cos(ωc * t) - 2 * a_c_y * sin(ωc * t)

    return [a_q, a_c]
end

function to_lab_frame(A::AbstractMatrix, ts::AbstractVector{<:Real})
    Ã = hcat([to_lab_frame(A[:, i], ts[i]) for i ∈ eachindex(ts)]...)
    return Ã
end

τs = [length(ts)]

lab_frame_experiment = QuantumSimulationExperiment(
    lab_frame_system,
    traj[1].ψ̃,
    ψ̃ -> ψ̃,
    τs;
    # drive_transform=A -> to_lab_frame(quantize_controls(A), ts)
    drive_transform=A -> quantize_controls(to_lab_frame(A, ts))
)

Ã = to_lab_frame(A, ts)

using CairoMakie

fig = Figure(resolution=(800, 600))
ax = Axis(fig[1, 1]; xlabel="t", ylabel="a")
series!(ax, ts, Ã; label="a_q_x")
fig

Ψ̃ = nominal_rollout(Ã, ts, lab_frame_experiment; integrator=exp)

pops = mapslices(populations, Ψ̃, dims=1)

fig2 = Figure(resolution=(800, 600))
ax2 = Axis(fig2[1, 1]; xlabel="t", ylabel="ψ")
series!(ax2, ts, pops[2:end, :], label="ψ", color=:prism)
fig2


populations(Ψ̃[:, end])

Ψ̃_quantized = rollout(A, ts, lab_frame_experiment)

population(Ψ̃_quantized[:, end], 1)

comps = (
    ψ̃ = Ψ̃,
    a = A,
    t = reshape(collect(ts), 1, :)
)

lab_frame_traj = NamedTrajectory(
    comps;
    timestep=dt,
    controls=:a,
    initial=(
        ψ̃ = Ψ̃[:, 1],
        a = A[:, 1],
        t = ts[1:1],
    ),
    final=(
        a = A[:, end],
        t = ts[end:end],
    ),
    bounds=(
        a = traj.bounds.a,
    )
)


function dynamics(
    zₜ::AbstractVector{<:Real},
    zₜ₊₁::AbstractVector{<:Real},
    Z::NamedTrajectory
)
    ψ̃ₜ = zₜ[Z.components.ψ̃]
    aₜ = zₜ[Z.components.a]
    tₜ = zₜ[Z.components.t][1]
    tₜ₊₁ = zₜ₊₁[Z.components.t][1]
    ãₜ = to_lab_frame(aₜ, tₜ)
    ψ̃ₜ₊₁ = zₜ₊₁[Z.components.ψ̃]
    Gₜ = QuantumCollocation.Integrators.G(ãₜ, lab_frame_system.G_drift, lab_frame_system.G_drives)
    return [ψ̃ₜ₊₁ - expv(Z.timestep, Gₜ, ψ̃ₜ); tₜ₊₁ - (tₜ + Z.timestep)]
end

P = FourthOrderPade(lab_frame_system)

function dynamics_pade(
    zₜ::AbstractVector{<:Real},
    zₜ₊₁::AbstractVector{<:Real},
    Z::NamedTrajectory
)
    ψ̃ₜ₊₁ = zₜ₊₁[Z.components.ψ̃]
    ψ̃ₜ = zₜ[Z.components.ψ̃]
    aₜ = zₜ[Z.components.a]
    ãₜ = to_lab_frame(aₜ, tₜ)
    tₜ₊₁ = zₜ₊₁[Z.components.t][1]
    tₜ = zₜ[Z.components.t][1]
    return [P(ψ̃ₜ₊₁, ψ̃ₜ, ãₜ, Z.timestep); tₜ₊₁ - (tₜ + Z.timestep)]
end



# f = zₜzₜ₊₁::AbstractVector{<:Real} -> begin
#     zₜ = zₜzₜ₊₁[slice(1, lab_frame_traj.dim)]
#     zₜ₊₁ = zₜzₜ₊₁[slice(2, lab_frame_traj.dim)]
#     return dynamics(zₜ, zₜ₊₁, lab_frame_traj)
# end

# f_pade = zₜzₜ₊₁::AbstractVector{<:Real} -> begin
#     zₜ = zₜzₜ₊₁[slice(1, lab_frame_traj.dim)]
#     zₜ₊₁ = zₜzₜ₊₁[slice(2, lab_frame_traj.dim)]
#     return dynamics_pade(zₜ, zₜ₊₁, lab_frame_traj)
# end

# f_exp = zₜzₜ₊₁::AbstractVector{<:Real} -> begin
#     zₜ = zₜzₜ₊₁[slice(1, lab_frame_traj.dim)]
#     zₜ₊₁ = zₜzₜ₊₁[slice(2, lab_frame_traj.dim)]
#     return dynamics_exp(zₜ, zₜ₊₁, lab_frame_traj)
# end

# using ForwardDiff, Zygote

# ∂f_forward_diff = zz -> ForwardDiff.jacobian(f, zz)
# ∂f_zygote = zz -> Zygote.jacobian(f, zz)[1]

# ∂f_pade_forward_diff = zz -> ForwardDiff.jacobian(f_pade, zz)
# ∂f_pade_zygote = zz -> Zygote.jacobian(f_pade, zz)[1]

# ∂f_exp_zygote = zz -> Zygote.jacobian(f_exp, zz)[1]

# # benchmark ∂f
# zz = vec(lab_frame_traj.data[:, 1000:1001])

# @benchmark ∂f_forward_diff($zz)
# @benchmark ∂f_zygote($zz)

# @benchmark ∂f_pade_forward_diff($zz)
# @benchmark ∂f_pade_zygote($zz)



max_iter = 20
max_backtrack_iter = 20
fps = 2
α = 0.5
β = 0.5^6
R = (a = 1.0e1,)
Qy = 1.0e-5
Qyf = 1.0e3
QP_tol = 1e-5
QP_max_iter = 1e9
QP_linear_solver = "mkl pardiso"

QP_settings = Dict(
    :polish => true,
)

static = false
nominal_correction = false
dynamic_measurements = true
correction_term = false

norm_p = 1

prob = ILCProblem(
    lab_frame_traj,
    dynamics,
    lab_frame_experiment;
    max_iter=max_iter,
    QP_verbose=false,
    correction_term=correction_term,
    dynamic_measurements=dynamic_measurements,
    norm_p=norm_p,
    R=R,
    static_QP=static,
    Qy=Qy,
    Qyf=Qyf,
    α=α,
    β=β,
    max_backtrack_iter=max_backtrack_iter,
    QP_tol=QP_tol,
    QP_max_iter=QP_max_iter,
    QP_settings=QP_settings,
    QP_linear_solver=QP_linear_solver,
    mle=true,
)

ILC.solve!(prob;
    nominal_correction=nominal_correction,
    show_metric=true,
    metric=:full_state,
)

plot_dir = joinpath(@__DIR__, "plots/lab_frame/")

plot_name =
    (static ? "static_" : "dynamic_") *
    "T_$(traj.T)_taus_$(length(τs))_" *
    "alpha_$(α)_beta_$(β)_" *
    "R_" * join([string(k) * "_" * string(v) for (k, v) in pairs(R)], "_") * "_" *
    "Qy_$(Qy)_Qyf_$(Qyf)_QP_tol_$(QP_tol)" *
    (nominal_correction ? "_nominal_correction" : "")

plot_path = generate_file_path("gif", plot_name, plot_dir)

animate_ILC_multimode(prob, plot_path; fps=fps, plot_transformed_controls=false)
