using CairoMakie
using PyCall
using LinearAlgebra

using QuantumCollocation
using NamedTrajectories
using IterativeLearningControl
const ILC = IterativeLearningControl


@pyinclude joinpath(@__DIR__, "lab_frame_quantization.py")

traj = load_traj(joinpath(@__DIR__, "../../data/multimode/g0_to_g1.jld2"))

ketdim = traj.dims.ψ̃ ÷ 2

println("dt = ", traj.timestep)
println()


scheme = "nearest_int"
# scheme = "abs"

function pops(ψ̃)
    ψ = iso_to_ket(ψ̃)
    return abs2.(ψ)
end

g_analytic = pops
# g_analytic = ψ̃ -> ψ̃
phase = exp(im * π / 2)

# Ψ̃_phased = mapslices(ket_to_iso, phase * mapslices(iso_to_ket, traj.ψ̃, dims=1), dims=1)

# traj.ψ̃ = Ψ̃_phased


function measure(controls, times, τs; scheme=scheme)
    ψf_py = py"lab_frame_experiment"(controls, times, τs, scheme=scheme)
    ψf = vec(vcat(ψf_py...))
    # ψf *= phase
    ψ̃f = ket_to_iso(ψf)
    y = g_analytic(ψ̃f)
    return [y]
end

y = measure(traj.a, times(traj), [traj.T]; scheme=scheme)[end]
println("|g1⟩ measured comp = ", iso_to_ket(y)[2])
println("|g1⟩ measured pop  = ", abs2(iso_to_ket(y)[2]))
println("|g1⟩ goal comp     = ", iso_to_ket(traj[end].ψ̃)[2])
println("|g1⟩ goal pop      = ", abs2(iso_to_ket(traj[end].ψ̃)[2]))

println()

# τs = [traj.T]
τs = 1:traj.T

transmon_levels = 3
cavity_levels = 14

system = MultiModeSystem(transmon_levels, cavity_levels, "g0", "g1")

P = FourthOrderPade(system)

function dynamics(
    zₜ::AbstractVector{<:Real},
    zₜ₊₁::AbstractVector{<:Real},
    Z::NamedTrajectory
)
    xₜ = zₜ[Z.components.states]
    xₜ₊₁ = zₜ₊₁[Z.components.states]
    uₜ = zₜ[Z.components.controls]
    Δt = Z.timestep
    return QuantumCollocation.Dynamics.dynamics(xₜ₊₁, xₜ, uₜ, Δt, P, system)
end

lab_frame_experiment = QuantumHardwareExperiment(
    measure,
    g_analytic,
    τs,
    traj[1].ψ̃
)

a_bounds = traj.bounds.a

function quantize(A::Matrix{Float64}; scheme=floor)
    res = a_bounds ./ 128
    Â = similar(A)
    for i = 1:size(A, 2)
        Â[:, i] = scheme.(A[:, i] ./ res) .* res
    end
    return Â
end

simple_experiment = QuantumSimulationExperiment(
    system,
    traj[1].ψ̃,
    g_analytic,
    τs;
    integrator=exp,
    control_transform=quantize,
)

max_iter = 20
max_backtrack_iter = 20
fps = 2
α = 0.5
β = 0.5^6
R = (
    a   = 1.0e1,
    dda = 1.0e10
)
Qy = 1.0e-5
Qyf = 1.0e3
QP_tol = 1e-6
QP_max_iter = 1e5
QP_linear_solver = "mkl pardiso"



QP_settings = Dict(
    :polish => true,
)

static = false
nominal_correction = false
dynamic_measurements = true
correction_term = true

norm_p = 1

# experiment = lab_frame_experiment
experiment = simple_experiment

y_final = [0; 1; zeros(ketdim - 2)]

prob = ILCProblem(
    traj,
    dynamics,
    experiment;
    max_iter=max_iter,
    QP_verbose=false,
    correction_term=correction_term,
    dynamic_measurements=dynamic_measurements,
    norm_p=norm_p,
    R=R,
    static_QP=static,
    Qy=Qy,
    Qyf=Qyf,
    α=α,
    β=β,
    max_backtrack_iter=max_backtrack_iter,
    QP_tol=QP_tol,
    QP_max_iter=QP_max_iter,
    QP_settings=QP_settings,
    QP_linear_solver=QP_linear_solver,
    mle=true,
    y_final=y_final,
)



ILC.solve!(prob;
    nominal_correction=nominal_correction,
    show_metric=true,
    # metric=:full_state,
)

A = prob.Ẑ.a

ψ̃f = measure(A, times(prob.Ẑ), τs; scheme=scheme)[1]

# display(norm(ψ̃f[[2, traj.dims.ψ̃ ÷ 2 + 2]])^2)
# display(fidelity(ψ̃f, traj[end].ψ̃))

plot_dir = joinpath(@__DIR__, "plots")

plot_name =
    (experiment == lab_frame_experiment ? "lab_frame_" : "simple_") *
    (static ? "static_" : "dynamic_") *
    "T_$(traj.T)_taus_$(length(τs))_" *
    "alpha_$(α)_beta_$(β)_" *
    "R_" * join([string(k) * "_" * string(v) for (k, v) in pairs(R)], "_") * "_" *
    "Qy_$(Qy)_Qyf_$(Qyf)_QP_tol_$(QP_tol)" *
    (nominal_correction ? "_nominal_correction" : "")

plot_path = generate_file_path("gif", plot_name, plot_dir)

animate_ILC_multimode(prob, plot_path; fps=fps, plot_transformed_controls=false)
