using Revise
using NamedTrajectories
using QuantumCollocation
using IterativeLearningControl
using LinearAlgebra
using CairoMakie

ILC_max_iter = 10
QP_max_iter = 50_000


data_path = joinpath(
    @__DIR__,
    "data/T_400_dt_0.1_Q_200.0_R_0.01_dda_bound_0.05_00001.jld2"
)

data = load_problem(data_path; return_data=true)

Z_ref = data["trajectory"]

integrators = data["integrators"]

system_ref = data["system"]

f(zₜ, zₜ₊₁) = vcat([integrator(zₜ, zₜ₊₁, Z_ref) for integrator ∈ integrators]...)

α = 1.0 * -2π * 0.143 #GHz
levels = 4

# drift_noise_term = 0.2 * create(levels) * create(levels) * annihilate(levels) * annihilate(levels)
drift_noise_term = 0.00007 * number(levels)
drive_1_noise_term = 0.01 * Hermitian(randn(ComplexF64, levels, levels))
drive_2_noise_term = 0.01 * Hermitian(randn(ComplexF64, levels, levels))

H_drift = α / 2 * quad(levels) + drift_noise_term
H_drives = [
    create(levels) + annihilate(levels),
    1im * (create(levels) - annihilate(levels))
]

experimental_system = QuantumSystem(
    H_drift,
    H_drives
)

U_iso_init = operator_to_iso_vec(I(levels))

ψ_init_g = cavity_state(0, levels)
ψ̃_init_g = ket_to_iso(ψ_init_g)

ψ_init_e = cavity_state(1, levels)
ψ̃_init_e = ket_to_iso(ψ_init_e)

ψ_goal_g = -1.0im * ψ_init_e
ψ̃_goal_g = ket_to_iso(ψ_goal_g)

ψ_goal_e = -1.0im * ψ_init_g
ψ̃_goal_e = ket_to_iso(ψ_goal_e)

ψ_init_ref_g = [1.0, 0.0]
ψ̃_init_ref_g = ket_to_iso(ψ_init_ref_g)

ψ_init_ref_e = [0.0, 1.0]
ψ̃_init_ref_e = ket_to_iso(ψ_init_ref_e)

ψ_goal_ref_g = -1.0im * ψ_init_ref_e
ψ̃_goal_ref_g = ket_to_iso(ψ_goal_ref_g)

ψ_goal_ref_e = -1.0im * ψ_init_ref_g
ψ̃_goal_ref_e = ket_to_iso(ψ_goal_ref_e)

U_goal_big = 2 .* [0.0 -1im 0 0;
         -1im 0. 0 0;
         0 0 0 0;
         0 0 0 0]

U_goal = [0. -1im ;
          -1im 0.]


X̃ = iso(GATES[:X])
Ỹ = iso(GATES[:Y])
Z̃ = iso(GATES[:Z])

measure(ψ̃::Vector, A::Matrix) = ψ̃' * A * ψ̃
# measure(Ũ::Matrix, A::Matrix) = 0.5 * trace(Ũ * A)

pauli(ψ̃::Vector) = [
    measure(ψ̃, X̃),
    measure(ψ̃, Ỹ),
    measure(ψ̃, Z̃)
]

# pauli(Ũ::Matrix) = [
#     measure(Ũ, X̃),
#     measure(Ũ, Ỹ),
#     measure(Ũ, Z̃)
# ]

g = ψ̃ -> pauli(ψ̃[[1, 2, levels+1, levels+2]])
g_ref = Ũ⃗ -> [
    pauli(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_g);
    pauli(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_e)
]

# g = Ũ⃗ -> [unitary_fidelity(iso_vec_to_operator(Ũ⃗), U_goal_big)]
# g_ref = Ũ⃗ -> [unitary_fidelity(iso_vec_to_operator(Ũ⃗), U_goal)]

# g = ψ̃ -> [fidelity(ψ̃, ψ̃_goal_g)]
# g_ref = Ũ⃗ -> [fidelity(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_g, ψ̃_goal_ref_g), fidelity(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_e, ψ̃_goal_ref_e)]

# g = ψ̃ -> populations(ψ̃)[1:2]
# g_ref = Ũ⃗ -> populations(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref)

# g = ψ̃ -> ψ̃[[1,2,5,6]]
# g_ref = Ũ⃗ -> iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref

τs = [Z_ref.T]

function quantize(A::Matrix{Float64}; scheme=floor)
    res = Z_ref.bounds.a[1] / 128
    Â = mapslices(a -> scheme.(a ./ res) .* res, A; dims=1)
    return Â
end

ψ̃_init = [ψ̃_init_g, ψ̃_init_e]

experiment = QuantumSimulationExperiment(experimental_system,
ψ̃_init,
#U_iso_init, 
g, τs;
    drive_transform=quantize,
    unitary = false,
)

y_goal = g_ref(Z_ref[end].Ũ⃗)

# J(x, Δts) = 0.5norm(experiment(x, Δts; only_final_measurement=true).ys[end] - y_goal, 2)^2
J(y_final) = norm(y_final - y_goal, 1)

α = 10.0
β = 0.001

optimizer = BacktrackingLineSearch(J; α=α, β=β, verbose=true)

R_value = 1e-4

prob = ILCProblem(Z_ref, system_ref, integrators, experiment, optimizer;
    g_ref=g_ref,
    max_iter=ILC_max_iter,
    QP_max_iter=QP_max_iter,
    R_value=R_value,
    tol=1e-6,
    cm_projection=true,
    cm_projection_tol=1e-5,
    isodim=4,
    unitary_ref_traj = true,
)

solve!(prob)

plot_path = generate_file_path("gif", "objective", "examples/transmon/plots/pi_gate/")

save_animation(plot_path, prob)
