using Revise
using NamedTrajectories
using QuantumCollocation
using IterativeLearningControl
using LinearAlgebra
using Suppressor
using PyCall

@pyinclude joinpath(@__DIR__(), "run_experiment_optimize_loop.py")

samples = 20000

function g(
    A::Matrix{Float64},
    times::AbstractVector{Float64},
    τs::AbstractVector{Int};
    acquisition_num::Int=samples
)
    as = collect(eachcol(A))
    ys = py"take_controls_and_measure"(times, as, τs, acquisition_num)
    ys = collect(eachcol(transpose(ys)))
    return ys
end



ILC_max_iter = 10
QP_max_iter = 50_000
stochastic = false
KL = false

data_path = joinpath(
    @__DIR__(),
    "data/T_400_dt_0.1_Q_200.0_R_0.01_dda_bound_0.05_00001.jld2"
)


data = load_problem(data_path; return_data=true)

Z_ref = data["trajectory"]

# display(g(Z_ref.a, times(Z_ref), [Z_ref.T]; acquisition_num=500))

integrators = data["integrators"]

α = 1.0 * -2π * 0.143 #GHz
levels = 4

ψ_init_g = cavity_state(0, levels)
ψ̃_init_g = ket_to_iso(ψ_init_g)

ψ_init_e = cavity_state(1, levels)
ψ̃_init_e = ket_to_iso(ψ_init_e)

ψ_goal_g = -1.0im * ψ_init_e
ψ̃_goal_g = ket_to_iso(ψ_goal_g)

ψ_goal_e = -1.0im * ψ_init_g
ψ̃_goal_e = ket_to_iso(ψ_goal_e)

ψ_init_ref_g = [1.0, 0.0]
ψ̃_init_ref_g = ket_to_iso(ψ_init_ref_g)

ψ_init_ref_e = [0.0, 1.0]
ψ̃_init_ref_e = ket_to_iso(ψ_init_ref_e)

ψ_goal_ref_g = -1.0im * ψ_init_ref_e
ψ̃_goal_ref_g = ket_to_iso(ψ_goal_ref_g)

ψ_goal_ref_e = -1.0im * ψ_init_ref_g
ψ̃_goal_ref_e = ket_to_iso(ψ_goal_ref_e)

X = GATES[:X]
Y = GATES[:Y]
Z = GATES[:Z]

measure(ψ̃::Vector, A::Matrix) =
    iso_to_ket(ψ̃)' * A * iso_to_ket(ψ̃) |> real

# measure(Ũ::Matrix, A::Matrix) = 0.5 * trace(Ũ * A)

pauli(ψ̃::Vector) = [
    (-measure(ψ̃, X) + 1) / 2,
    (-measure(ψ̃, Y) + 1) / 2,
    (-measure(ψ̃, Z) + 1) / 2
]

acquisition_num = 100_000


g_ref = Ũ⃗ -> [
    pauli(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_g);
    pauli(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_e)
]

# display(g_ref(Z_ref[end].Ũ⃗))

ydim = 6

τs = [Z_ref.T]

ψ̃_init = [ψ̃_init_g, ψ̃_init_e]

experiment = QuantumHardwareExperiment(g, ydim, τs; stochastic=stochastic)

y_goal = g_ref(Z_ref[end].Ũ⃗)

J(y_final) = norm(y_final - y_goal, 1)

α = 1.0
β = 0.001
optimizer = BacktrackingLineSearch(J; α=α, β=β, verbose=true, global_min=true)

# α = 1e-3
# β₁ = 0.75
# optimizer = ADAM(J, Z_ref; α=α, β₁=β₁)

R_value = 1e-4

prob = ILCProblem(Z_ref, integrators, experiment, optimizer;
    g_ref=g_ref,
    max_iter=ILC_max_iter,
    QP_max_iter=QP_max_iter,
    R_value=R_value,
    KL=KL,
    Qyf=KL ? 100.0 : 100.0,
)

experiment_name = "samples_$(samples)"

io = open(generate_file_path(
    "txt",
    experiment_name * "_log",
    @__DIR__() * "//pi_gate_hardware"
), "w")

solve!(prob; io=io)


plot_path = generate_file_path(
    "gif",
    experiment_name * "_objective",
    @__DIR__() * "//pi_gate_hardware"
)

save_animation(plot_path, prob)

cos_sims_plot_path = generate_file_path(
    "png",
    experiment_name * "_cos_sims",
    @__DIR__() * "//pi_gate_hardware"
)

dZs = [dZ.datavec for dZ ∈ prob.dZs]

plot_cosine_similarity(cos_sims_plot_path, dZs)
