using Revise
using NamedTrajectories
using QuantumCollocation
using IterativeLearningControl
using LinearAlgebra
using CairoMakie

ILC_max_iter = 20
QP_max_iter = 50_000
KL = true

data_path = joinpath(
    @__DIR__,
    "data/T_400_dt_0.1_Q_200.0_R_0.01_dda_bound_0.05_00001.jld2"
)


data = load_problem(data_path; return_data=true)

Z_ref = data["trajectory"]
system_ref = data["system"]

integrators = data["integrators"]

f(zₜ, zₜ₊₁) = vcat([integrator(zₜ, zₜ₊₁, Z_ref) for integrator ∈ integrators]...)

α = 1.0 * -2π * 0.143 #GHz
levels = 4

# drift_noise_term = 0.2 * create(levels) * create(levels) * annihilate(levels) * annihilate(levels)
drift_noise_term = 0.000001 * number(levels)
drive_1_noise_term = 0.01 * Hermitian(randn(ComplexF64, levels, levels))
drive_2_noise_term = 0.01 * Hermitian(randn(ComplexF64, levels, levels))

H_drift = α / 2 * quad(levels) + drift_noise_term
H_drives = [
    create(levels) + annihilate(levels),
    1im * (create(levels) - annihilate(levels))
]

experimental_system = QuantumSystem(
    H_drift,
    H_drives
)

ψ_init_g = cavity_state(0, levels)
ψ̃_init_g = ket_to_iso(ψ_init_g)

ψ_init_e = cavity_state(1, levels)
ψ̃_init_e = ket_to_iso(ψ_init_e)

ψ_goal_g = -1.0im * ψ_init_e
ψ̃_goal_g = ket_to_iso(ψ_goal_g)

ψ_goal_e = -1.0im * ψ_init_g
ψ̃_goal_e = ket_to_iso(ψ_goal_e)

ψ_init_ref_g = [1.0, 0.0]
ψ̃_init_ref_g = ket_to_iso(ψ_init_ref_g)

ψ_init_ref_e = [0.0, 1.0]
ψ̃_init_ref_e = ket_to_iso(ψ_init_ref_e)

ψ_goal_ref_g = -1.0im * ψ_init_ref_e
ψ̃_goal_ref_g = ket_to_iso(ψ_goal_ref_g)

ψ_goal_ref_e = -1.0im * ψ_init_ref_g
ψ̃_goal_ref_e = ket_to_iso(ψ_goal_ref_e)

X = GATES[:X]
Y = GATES[:Y]
Z = GATES[:Z]

function iso_operator_to_operator(A)
    n = size(A, 1)
    A_real = A[1:n, 1:n]
    A_imag = A[n+1:end, 1:n]
    return A_real + 1.0im * A_imag
end

measure(ψ̃::Vector, A::Matrix) =
    iso_to_ket(ψ̃)' * A * iso_to_ket(ψ̃) |> real

# measure(Ũ::Matrix, A::Matrix) = 0.5 * trace(Ũ * A)

pauli(ψ̃::Vector) = [
    measure(ψ̃, X),
    measure(ψ̃, Y),
    measure(ψ̃, Z)
]

# pauli(Ũ::Matrix) = [
#     measure(Ũ, X̃),
#     measure(Ũ, Ỹ),
#     measure(Ũ, Z̃)
# ]

acquisition_num = 20_000

g = ψ̃ -> pauli_meas_g(
    iso_to_ket(ψ̃[[1, 2, levels+1, levels+2]]);
    return_variances=true,
    acquisition_num=acquisition_num
)

g_ref = Ũ⃗ -> [
    pauli(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_g);
    pauli(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref_e)
]

# g = ψ̃ -> [fidelity(ψ̃, ψ̃_goal)]
# g_ref = Ũ⃗ -> [fidelity(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref, ψ̃_goal_ref)]

# g = ψ̃ -> populations(ψ̃)[1:2]
# g_ref = Ũ⃗ -> populations(iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref)

# g = ψ̃ -> ψ̃[[1,2,5,6]]
# g_ref = Ũ⃗ -> iso_vec_to_iso_operator(Ũ⃗) * ψ̃_init_ref

τs = [Z_ref.T]

function quantize(A::Matrix{Float64}; scheme=floor)
    res = Z_ref.bounds.a[1] / 128
    Â = mapslices(a -> scheme.(a ./ res) .* res, A; dims=1)
    return Â
end

ψ̃_init = [ψ̃_init_g, ψ̃_init_e]

experiment = QuantumSimulationExperiment(experimental_system, ψ̃_init, g, τs;
    # drive_transform=quantize,
    stochastic=true,
)

y_goal = g_ref(Z_ref[end].Ũ⃗)

J(y_final) = norm(y_final - y_goal, 1)

α = 1.0
β = 0.0
optimizer = BacktrackingLineSearch(J; α=α, β=β, verbose=true, global_min=true)

# α = 1e-3
# β₁ = 0.75
# optimizer = ADAM(J, Z_ref; α=α, β₁=β₁)

# θ = 0.0
# optimizer = StochasticLineSearch(J; θ=θ, verbose=true)

R_value = 1e-4

prob = ILCProblem(Z_ref, system_ref, integrators, experiment, optimizer;
    g_ref=g_ref,
    max_iter=ILC_max_iter,
    OSQP_max_iter=QP_max_iter,
    R_value=R_value,
    KL=KL,
    Qyf=KL ? 100.0 : 100.0,
)

io = open(joinpath(@__DIR__, "log.txt"), "w")
# io = stdout

solve!(prob; io=io)

plot_path = generate_file_path(
    "gif",
    "objective",
    "examples/transmon/plots/pi_gate_noisy/"
)

save_animation(plot_path, prob)

cos_sims_plot_path = generate_file_path(
    "png",
    "cos_sims",
    "examples/transmon/plots/pi_gate_noisy/"
)

dZs = [dZ.datavec for dZ ∈ prob.dZs]

plot_cosine_similarity(cos_sims_plot_path, dZs)
