module Experiments

export AbstractExperiment

export QuantumSimulationExperiment

export QuantumHardwareExperiment

export rollout
export nominal_rollout

using QuantumCollocation


using LinearAlgebra
using BasicInterpolators

using QuantumOptics

using ..Measurements
using ..MeasurementFunctions


abstract type AbstractExperiment end


struct QuantumSimulationExperiment <: AbstractExperiment
    x_init::Union{AbstractVector{Float64}, Vector{<:AbstractVector{Float64}}}
    g::Union{Function,Vector{Function}}
    stochastic::Bool
    ydim::Int
    τs::AbstractVector{Int}
    M::Int
    integrator::Function
    drive_transform::Function
    G_drift::AbstractMatrix{Float64}
    G_drives::Vector{AbstractMatrix{Float64}}
    G_error_term::AbstractMatrix{Float64}
    unitary::Bool
end

function QuantumSimulationExperiment(
    sys::QuantumCollocation.QuantumSystems.QuantumSystem,
    x_init::AbstractVector{Float64},
    g::Function,
    τs::AbstractVector{Int};
    stochastic=false,
    G_error_term=zeros(size(sys.G_drift)),
    integrator=exp,
    drive_transform=A->A,
    unitary=false
)
    if stochastic
        ydim = length(g(x_init)[1])
    else
        ydim = length(g(x_init))
    end
    @assert 1 ∉ τs "τs must not contain initial time step"

    return QuantumSimulationExperiment(
        x_init,
        g,
        stochastic,
        ydim,
        τs,
        length(τs),
        integrator,
        drive_transform,
        sys.G_drift,
        sys.G_drives,
        G_error_term,
        unitary
    )
end

function QuantumSimulationExperiment(
    sys::QuantumCollocation.QuantumSystems.QuantumSystem,
    x_inits::Vector{<:AbstractVector{Float64}},
    g::Function,
    τs::AbstractVector{Int};
    stochastic=false,
    G_error_term=zeros(size(sys.G_drift)),
    integrator=exp,
    drive_transform=A->A,
    unitary=false
)
    if stochastic
        @assert g(x_inits[1]) isa Tuple
        ydim = length(g(x_inits[1])[1]) * length(x_inits)
    else
        ydim = length(g(x_inits[1])) * length(x_inits)
    end
    return QuantumSimulationExperiment(
        x_inits,
        g,
        stochastic,
        ydim,
        τs,
        length(τs),
        integrator,
        drive_transform,
        sys.G_drift,
        sys.G_drives,
        G_error_term,
        unitary
    )
end


function QuantumSimulationExperiment(
    sys::QuantumCollocation.QuantumSystems.QuantumSystem,
    x_inits::Vector{<:AbstractVector{Float64}},
    gs::Vector{Function},
    τs::AbstractVector{Int};
    stochastic=false,
    G_error_term=zeros(size(sys.G_drift)),
    integrator=exp,
    drive_transform=A->A,
    unitary=false
)
    @assert length(x_inits) == length(gs)
    if stochastic
        # TODO: clean up this case
        @assert gs[1](x_inits[1]) isa Tuple
        ydim = length(gs[1](x_inits[1])[1]) * length(x_inits)
    else
        ydim = sum([length(g(x_init)) for (g, x_init) in zip(gs, x_inits)])
    end
    return QuantumSimulationExperiment(
        x_inits,
        gs,
        stochastic,
        ydim,
        τs,
        length(τs),
        integrator,
        drive_transform,
        sys.G_drift,
        sys.G_drives,
        G_error_term,
        unitary
    )
end

function (experiment::QuantumSimulationExperiment)(
    A::AbstractMatrix{Float64},
    Δts::AbstractVecOrMat{Float64};
    only_final_measurement=false,
)
    if Δts isa AbstractMatrix
        @assert size(Δts, 1) == 1
        Δts = Δts[1, :]
    end

    if experiment.stochastic
        Measurement = StochasticMeasurement
    else
        Measurement = DeterministicMeasurement
    end

    if experiment.unitary
        X = unitary_rollout(experiment.x_init, A, Δts, experiment)
    else
        X = quantum_state_rollout(experiment.x_init, A, Δts, experiment)
    end

    if only_final_measurement
        Y = Measurement(
            X,
            experiment.g,
            experiment.τs[end:end],
            experiment.ydim
        )
    else
        Y = Measurement(
            X,
            experiment.g,
            experiment.τs,
            experiment.ydim
        )
    end

    return Y
end

function quantum_state_rollout(
    ψ̃_init::AbstractVector{Float64},
    A::AbstractMatrix{Float64},
    Δt::AbstractVector{Float64},
    experiment::QuantumSimulationExperiment
)
    T = size(A, 2)

    Ψ̃ = zeros(length(ψ̃_init), T)

    Ψ̃[:, 1] .= ψ̃_init

    Â = experiment.drive_transform(A)

    for t = 2:T
        âₜ₋₁ = Â[:, t - 1]
        Gₜ = QuantumCollocation.Integrators.G(
            âₜ₋₁,
            experiment.G_drift,
            experiment.G_drives
        ) + experiment.G_error_term
        Ψ̃[:, t] .= experiment.integrator(Gₜ * Δt[t - 1]) * Ψ̃[:, t - 1]
    end

    return Ψ̃
end

function unitary_rollout(
    Ũ⃗_init::AbstractVector{<:Real},
    A::AbstractMatrix{Float64},
    Δt::Union{AbstractVector{Float64}, AbstractMatrix{Float64}},
    experiment::QuantumSimulationExperiment
)
    if Δt isa AbstractMatrix
        @assert size(Δt, 1) == 1
        Δt = vec(Δt)
    end

    T = size(A, 2)

    Ũ⃗ = zeros(length(Ũ⃗_init), T)

    Ũ⃗[:, 1] .= Ũ⃗_init

    Â = experiment.drive_transform(A)

    for t = 2:T
        âₜ₋₁ = Â[:, t - 1]
        Gₜ = QuantumCollocation.Integrators.G(
            âₜ₋₁,
            experiment.G_drift,
            experiment.G_drives
        ) + experiment.G_error_term
        Ũ⃗ₜ₋₁ = Ũ⃗[:, t - 1]
        Ũₜ₋₁ = iso_vec_to_iso_operator(Ũ⃗ₜ₋₁)
        Ũₜ = experiment.integrator(Gₜ * Δt[t - 1]) * Ũₜ₋₁
        Ũ⃗ₜ = iso_operator_to_iso_vec(Ũₜ)
        Ũ⃗[:, t] .= Ũ⃗ₜ
    end

    return Ũ⃗
end

quantum_state_rollout(ψ̃_inits::Vector{<:AbstractVector{Float64}}, args...) =
    [quantum_state_rollout(ψ̃, args...) for ψ̃ ∈ ψ̃_inits]

function nominal_rollout(
    ψ̃_init::AbstractVector{Float64},
    A::AbstractMatrix{Float64},
    Δt::AbstractVector{Float64},
    experiment::QuantumSimulationExperiment;
    integrator=QuantumCollocation.Integrators.fourth_order_pade
)
    T = size(A, 2)

    Ψ̃ = zeros(length(ψ̃_init), T)

    Ψ̃[:, 1] .= ψ̃_init

    for t = 2:T
        aₜ₋₁ = A[:, t - 1]
        Gₜ = Integrators.G(
            aₜ₋₁,
            experiment.G_drift,
            experiment.G_drives
        )
        Ψ̃[:, t] .= integrator(Gₜ * Δt[t - 1]) * Ψ̃[:, t - 1]
    end

    return Ψ̃
end

nominal_rollout(ψ̃_inits::Vector{<:AbstractVector{Float64}}, args...) =
    [nominal_rollout(ψ̃ⁱ, args...) for ψ̃ⁱ in ψ̃_inits]



end
