module ILCProblems

export ILCProblem
export save

using ..Measurements
using ..Experiments
using ..Optimizers
using ..QuadraticProblems

using TrajectoryIndexingUtils
using NamedTrajectories
using QuantumCollocation

using LinearAlgebra
using SparseArrays
using Statistics
using JLD2



mutable struct ILCProblem
    experiment::AbstractExperiment
    optimizer::AbstractOptimizer
    system_ref::QuantumSystem
    unitary_ref_traj::Bool
    QP::QuadraticProblem
    state_names::Vector{Symbol}
    isodim::Union{Int, Nothing}
    drive_name::Symbol
    drive_deriv_name::Symbol
    continuous_traj::Bool
    Z_goal::NamedTrajectory
    Zs::Vector{NamedTrajectory}
    dZs::Vector{NamedTrajectory}
    Y_goal::DeterministicMeasurement
    Ys::Vector{AbstractMeasurement}
    objective_history::Vector{Float64}
    settings::Dict{Symbol, Any}
end

"""
    save(file::String, prob::ILCProblem)

    Save all data of an ILCProblem (not including components containing functions) to a file.
"""
function JLD2.save(file::String, prob::ILCProblem; kwargs...)
    data = Dict(
        "system_ref" => prob.system_ref,
        "unitary_ref_traj" => prob.unitary_ref_traj,
        "state_names" => prob.state_names,
        "isodim" => prob.isodim,
        "drive_name" => prob.drive_name,
        "drive_deriv_name" => prob.drive_deriv_name,
        "continuous_traj" => prob.continuous_traj,
        "Z_goal" => prob.Z_goal,
        "Zs" => prob.Zs,
        "dZs" => prob.dZs,
        "Y_goal" => prob.Y_goal,
        "Ys" => prob.Ys,
        "objective_history" => prob.objective_history,
        "settings" => prob.settings
    )
    # join with kwargs
    data = merge(data, Dict{String, Any}(kwargs...))
    save(file, data)
end


function ILCProblem(
    Z_goal::NamedTrajectory,
    system_ref::QuantumSystem,
    integrators::Vector{AbstractIntegrator},
    experiment::AbstractExperiment,
    optimizer::AbstractOptimizer;
    g_ref::Union{Function, Vector{Function}}=
        experiment.stochastic ? x -> experiment.g(x)[1] : experiment.g,
    quantum_state_name::Symbol=:Ũ⃗,
    quantum_state_names::Vector{Symbol}=[quantum_state_name],
    drive_name::Symbol=:a,
    drive_deriv_name::Symbol=:da,
    continuous_traj::Bool=false,
    Q=1.0,
    Qf=100.0,
    R_value=0.1,
    R=merge(
        (a=0.1R_value, da=R_value, dda=2.0R_value),
        NamedTuple((quantum_state_name => 0.01R_value,))
    ),
    objective_hessian=false,
    max_iter=100,
    tol=1e-2,
    norm_p=Inf,
    QP_verbose=false,
    OSQP_max_iter=20_000,
    OSQP_verbose=false,
    OSQP_linear_solver="qdldl",
    OSQP_settings=Dict(),
    OSQP_tol=1e-2,
    OSQP_polish=true,
    KL=true,
    y_final=nothing,
    verbose=true,
    cm_projection=false,
    cm_projection_tol=1e-4,
    isodim=nothing,
    unitary_ref_traj=true,
    io::IO=stdout,
)
    @assert R isa NamedTuple "R must be a NamedTuple specifying components to regularize -- e.g. R=(a=1.0, dda=1.0)"

    @assert all(name ∈ Z_goal.names for name in quantum_state_names) "state_names must be a subset of the names of Z_goal"

    if g_ref isa Function
        X = vcat([Z_goal[name] for name ∈ quantum_state_names]...)

        @assert g_ref(X[:, end]) isa AbstractVector{Float64} "g_ref must be a deterministic measurement function"

        Y_goal = DeterministicMeasurement(
            X,
            g_ref,
            experiment.τs,
            experiment.ydim
        )
    elseif g_ref isa Vector{Function}
        Xs = [Z_goal[name] for name ∈ quantum_state_names]

        @assert length(g_ref) == length(quantum_state_names) "number of functions in g_ref must match number of quantum_state_names"

        @assert all(g(X[:, end]) isa AbstractVector{Float64} for (g, X) ∈ zip(g_ref, Xs)) "g_ref must be a vector of deterministic measurement functions"

        Y_goal = DeterministicMeasurement(
            Xs,
            g_ref,
            experiment.τs,
            experiment.ydim
        )
    else
        @error "g_ref must be a function or a vector of functions"
    end

    ILC_settings::Dict{Symbol, Any} = Dict(
        :max_iter => max_iter,
        :tol => tol,
        :cm_proj => cm_projection,
        :cm_proj_tol => cm_projection_tol,
        :norm_p => norm_p,
        :verbose => verbose,
    )

    OSQP_kw_settings::Dict{Symbol, Any} = Dict(
        :max_iter => OSQP_max_iter,
        :verbose => OSQP_verbose,
        :eps_abs => OSQP_tol,
        :eps_rel => OSQP_tol,
        :eps_prim_inf => OSQP_tol,
        :eps_dual_inf => OSQP_tol,
        :linsys_solver => OSQP_linear_solver,
        :polish => OSQP_polish
    )

    OSQP_settings::Dict{Symbol, Any} =
        merge(OSQP_kw_settings, OSQP_settings)

    println(io, "Setting up dynamic QP...")

    QP = DynamicQuadraticProblem(
        Z_goal,
        quantum_state_names,
        integrators,
        g_ref,
        experiment.τs,
        Q, Qf, R,
        objective_hessian,
        KL,
        QP_verbose,
        io,
        OSQP_settings
    )

    if !isnothing(y_final)
        Y_goal.ys[end] = y_final
    end

    return ILCProblem(
        experiment,
        optimizer,
        system_ref,
        unitary_ref_traj,
        QP,
        quantum_state_names,
        isodim,
        drive_name,
        drive_deriv_name,
        continuous_traj,
        Z_goal,
        NamedTrajectory[Z_goal],
        NamedTrajectory[],
        Y_goal,
        AbstractMeasurement[],
        Float64[],
        ILC_settings,
    )
end

function ILCProblem(
    prob::QuantumControlProblem,
    experiment::AbstractExperiment;
    unitary_ref_traj=true,
    quantum_state_name=:Ũ⃗,
    quantum_state_names=[quantum_state_name],
    α=0.5,
    α_cutoff=α * 0.5^10,
    g_ref=experiment.stochastic ? x -> experiment.g(x)[1] : experiment.g,
    optimizer::Union{Nothing, AbstractOptimizer}=nothing,
    optimizer_verbose=true,
    kwargs...
)
    if isnothing(optimizer)
        if unitary_ref_traj
            y_goal = g_ref(prob.trajectory[quantum_state_name][:, end])
        else
            @error "non-unitary reference trajectories not yet implemented for this method"
        end
        J(y) = norm(y - y_goal, 1)
        optimizer = BacktrackingLineSearch(J; α=α, cutoff=α_cutoff, verbose=optimizer_verbose)
    end

    return ILCProblem(
        prob.trajectory,
        prob.system,
        prob.integrators,
        experiment,
        optimizer;
        g_ref=g_ref,
        quantum_state_names=quantum_state_names,
        unitary_ref_traj=unitary_ref_traj,
        kwargs...
    )
end


end
