module MeasurementFunctions

export meas_x
export meas_x_iso
export meas_x_noisy
export meas_x_noisy_iso

export meas_y
export meas_y_iso
export meas_y_noisy
export meas_y_noisy_iso

export meas_z
export meas_z_iso
export meas_z_noisy
export meas_z_noisy_iso

export pauli_meas_g
export pauli_meas_g_iso

using LinearAlgebra
using Random
using Distributions
using Statistics



#
# Noisy Operators and Measurements
#
function get_mat_iso(mat)
    mat_r = real(mat)
    mat_i = imag(mat)
    return vcat(hcat(mat_r, -mat_i),
                hcat(mat_i,  mat_r))
end

function get_mat_uniso(mat)
    n = size(mat, 1)
    nby2 = Integer(n/2)
    i1 = 1:nby2
    i2 = (nby2 + 1):n
    return mat[i1, i1] + mat[i2, i1]
end

function sigmaz()
    return [1 0;
            0 -1]
end

function sigmax()
    return [0 1;
            1 0]
end

function sigmay()
    return [0 -1im;
            1im 0]
end

function noisy_sigmax(;error = 0.025)
    H = noisy_hadamard(error = error)
    return H*sigmaz()*H
end

function noisy_sigmay(; error = 0.025)
    HS_dag = noisy_hadamard(error = error) * S_dag()
    return HS_dag' * sigmaz() * HS_dag
end

function S_dag()
    return [1 0;
            0 -1im]
end

function noisy_hadamard(;error = 0.025)
    d = Normal(π/2, π/2 * error)
    θ =  rand(d)
    return [cos(θ/2) sin(θ/2);
            sin(θ/2) -cos(θ/2)]
end

function meas_x(state::Vector{<:Number})
    real(state' * sigmax() * state)
end

function meas_x_iso(iso_state::Vector{<:Real})
    iso_state' * get_mat_iso(sigmax()) * iso_state
end

function meas_y(state::Vector{<:Number})
    real(state' * sigmay() * state)
end

function meas_y_iso(iso_state::Vector{<:Real})
    iso_state' * get_mat_iso(sigmay()) * iso_state
end

function meas_x_noisy(state::Vector{<:Number}; error = 0.025)
    #might need to take real part cause of numerical error
    real(state' * noisy_sigmax(error = error) * state)
end

function meas_x_noisy_iso(iso_state::Vector{<:Real}; error = 0.025)
    #might need to take real part cause of numerical error
    iso_state' * get_mat_iso(noisy_sigmax(error = error)) * iso_state
end


function meas_y_noisy(state::Vector{<:Number}; error = 0.025)
    real(state' * noisy_sigmay(error = error) * state)
end

function meas_y_noisy_iso(iso_state::Vector{<:Real}; error = 0.025)
    #might need to take real part cause of numerical error
    iso_state' * get_mat_iso(noisy_sigmay(error = error)) * iso_state
end

function meas_z(state::Vector{<:Number})
    real(state' * sigmaz() * state)
end

function meas_z_iso(iso_state::Vector{<:Real})
    iso_state' * get_mat_iso(sigmaz()) * iso_state
end


function pauli_meas_g(
    state::Vector{<:Number};
    return_variances = false,
    error = 0.025,
    acquisition_num = 1000,
    readout_fidelity = 0.96
)
    σxs = []
    σys = []
    σzs = []

    for _ = 1:acquisition_num
        push!(σxs, sign(
            rand(Uniform(-1, 1)) +
            (1 - 2 * (1 - readout_fidelity)) * meas_x_noisy(state, error = error)
        ))
        push!(σys, sign(
            rand(Uniform(-1, 1)) +
            (1 - 2 * (1 - readout_fidelity)) * meas_y_noisy(state, error = error)
        ))
        push!(σzs, sign(
            rand(Uniform(-1, 1)) +
            (1 - 2 * (1 - readout_fidelity)) * meas_z(state)
        ))
    end

    μs = [mean(σxs), mean(σys), mean(σzs)]

    if return_variances
        vs = [
            var(σxs) / acquisition_num,
            var(σys) / acquisition_num,
            var(σzs) / acquisition_num
        ]
        return μs, vs
    else
        return μs
    end
end

function pauli_meas_g_iso(
    iso_state::Vector{<:Real};
    error = 0.025,
    acquisition_num = 1000,
    readout_fidelity = 0.96
)
    running_x = 0
    running_y = 0
    running_z = 0

    for _ = 1:acquisition_num
        running_x += sign(
            rand() -
            (1 - readout_fidelity) *
            meas_x_noisy_iso(iso_state, error = error)
        )

        running_y += sign(
            rand() -
            (1 - readout_fidelity) *
            meas_y_noisy_iso(iso_state, error = error)
        )

        running_z += sign(
            rand() -
            (1 - readout_fidelity) *
            meas_z_iso(iso_state)
        )
    end

    return [
        running_x / acquisition_num,
        running_y / acquisition_num,
        running_z / acquisition_num
    ]
end

end
