module Measurements

export AbstractMeasurement
export DeterministicMeasurement
export StochasticMeasurement

using LinearAlgebra


abstract type AbstractMeasurement end

function LinearAlgebra.norm(data::AbstractMeasurement, p::Real=2)
    return norm(vcat(data.ys...), p)
end

struct DeterministicMeasurement <: AbstractMeasurement
    ys::Vector{<:AbstractVector{Float64}}
    τs::AbstractVector{Int}
    ydim::Int
end

function DeterministicMeasurement(
    X::AbstractMatrix{Float64},
    g::Function,
    τs::AbstractVector{Int},
    ydim::Int
)
    @assert length(g(X[:, end])) == ydim "length(g(x)) = $(length(g(X[:, 1]))) != ydim = $(ydim))"

    ys = Vector{Vector{Float64}}(undef, length(τs))

    for (i, τ) in enumerate(τs)
        ys[i] = g(X[:, τ])
    end

    return DeterministicMeasurement(ys, τs, ydim)
end

function DeterministicMeasurement(
    Xs::Vector{<:AbstractMatrix{Float64}},
    g::Function,
    τs::AbstractVector{Int},
    ydim::Int
)
    @assert length(g(Xs[1][:, 1])) * length(Xs) == ydim

    ys = Vector{Vector{Float64}}(undef, length(τs))

    for (i, τ) ∈ enumerate(τs)
        ys_i = []
        for X ∈ Xs
            append!(ys_i, g(X[:, τ]))
        end
        ys[i] = ys_i
    end

    return DeterministicMeasurement(ys, τs, ydim)
end

function DeterministicMeasurement(
    Xs::Vector{<:AbstractMatrix{Float64}},
    gs::Vector{Function},
    τs::AbstractVector{Int},
    ydim::Int
)
    @assert length(gs) == length(Xs)
    @assert ydim == sum([length(g(Xs[1][:, 1])) for g ∈ gs])

    ys = Vector{Vector{Float64}}(undef, length(τs))

    for (i, τ) ∈ enumerate(τs)
        ys_i = []
        for (g, X) ∈ zip(gs, Xs)
            append!(ys_i, g(X[:, τ]))
        end
        ys[i] = ys_i
    end

    return DeterministicMeasurement(ys, τs, ydim)
end






struct StochasticMeasurement <: AbstractMeasurement
    ys::Vector{<:AbstractVector{Float64}}
    Σs::Vector{<:AbstractMatrix{Float64}}
    τs::AbstractVector{Int}
    ydim::Int
end

function StochasticMeasurement(
    X::AbstractMatrix{Float64},
    g::Function,
    τs::AbstractVector{Int},
    ydim::Int
)
    @assert length(g(X[:, 1])) == ydim "length(g(x)) = $(length(g(X[:, 1]))) != ydim = $(ydim)"

    ys = Vector{Vector{Float64}}(undef, length(τs))
    Σs = Vector{Matrix{Float64}}(undef, length(τs))

    for (i, τ) ∈ enumerate(τs)
        yᵢ, σᵢ = g(X[:, τ])
        ys[i] = yᵢ
        Σs[i] = diagm(σᵢ)
    end

    return StochasticMeasurement(ys, Σs, τs, ydim)
end

function StochasticMeasurement(
    Xs::Vector{<:AbstractMatrix{Float64}},
    g::Function,
    τs::AbstractVector{Int},
    ydim::Int
)
    @assert length(g(Xs[1][:, 1])[1]) * length(Xs) == ydim

    ys = Vector{Vector{Float64}}(undef, length(τs))
    Σs = Vector{Matrix{Float64}}(undef, length(τs))

    for (i, τ) ∈ enumerate(τs)
        yᵢ = Float64[]
        σᵢ = Float64[]
        for Xⱼ ∈ Xs
            yᵢⱼ, σᵢⱼ = g(Xⱼ[:, τ])
            append!(yᵢ, yᵢⱼ)
            append!(σᵢ, σᵢⱼ)
        end
        ys[i] = yᵢ
        if σᵢ isa Vector
            Σs[i] = diagm(σᵢ)
        else
            Σs[i] = σᵢ
        end
    end

    return StochasticMeasurement(ys, Σs, τs, ydim)
end

function Base.:-(data1::DeterministicMeasurement, data2::DeterministicMeasurement)
    @assert data1.ydim == data2.ydim
    @assert data1.τs == data2.τs
    return DeterministicMeasurement(
        data1.ys .- data2.ys,
        data1.τs,
        data1.ydim
    )
end

function Base.:-(data1::StochasticMeasurement, data2::StochasticMeasurement)
    @assert data1.ydim == data2.ydim
    @assert data1.τs == data2.τs
    return StochasticMeasurement(
        data1.ys .- data2.ys,
        data1.Σs .+ data2.Σs, # check this
        data1.τs,
        data1.ydim
    )
end

function Base.:-(m1::DeterministicMeasurement, m2::StochasticMeasurement)
    @assert m1.ydim == m2.ydim
    @assert m1.τs == m2.τs
    return StochasticMeasurement(
        m1.ys .- m2.ys,
        m2.Σs,
        m1.τs,
        m1.ydim
    )
end

function Base.:-(m1::StochasticMeasurement, m2::DeterministicMeasurement)
    @assert m1.ydim == m2.ydim
    @assert m1.τs == m2.τs
    return StochasticMeasurement(
        m1.ys .- m2.ys,
        m1.Σs,
        m1.τs,
        m1.ydim
    )
end



end
