module Optimizers

export AbstractOptimizer
export BacktrackingLineSearch

using ..Experiments

using QuantumCollocation.QuantumUtils
using NamedTrajectories
using LinearAlgebra
using Manifolds

abstract type AbstractOptimizer end

"""

"""
mutable struct BacktrackingLineSearch <: AbstractOptimizer
    J::Function
    J_min::Float64
    # ∇J::Function # TODO: derive from f
    α::Float64
    p::Float64
    β::Float64
    cutoff::Float64
    global_min::Bool
    verbose::Bool
end

function BacktrackingLineSearch(
    J::Function;
    α::Float64=1.0,
    p::Float64=0.5,
    β::Float64=1e-4,
    cutoff::Float64=0.001,
    global_min=false,
    verbose=false
)
    return BacktrackingLineSearch(J, Inf, α, p, β, cutoff, global_min, verbose)
end

function (opt::BacktrackingLineSearch)(
    Z::NamedTrajectory,
    dZ::NamedTrajectory,
    J₀::Float64,
    drive_name::Symbol,
    experiment::AbstractExperiment;
    io::IO=stdout,
    manifold_projection=true,
    unitary=true,
    kwargs...
)::NamedTrajectory
    α = opt.α
    p = opt.p
    β = opt.β

    A = Z[drive_name]
    dA = dZ[drive_name]

    Δt = timesteps(Z)
    dΔt = timesteps(dZ)

    d² = dot(vcat(vec(dA), vec(dΔt)), vcat(vec(dA), vec(dΔt)))

    y_final = experiment(
        A + α * dA,
        Δt + α * dΔt;
        only_final_measurement=true
    ).ys[end]

    J = opt.J(y_final)

    if opt.verbose
        println(io, "        J = $J, α = $α")
        flush(io)
    end

    if opt.global_min
        J₀ = minimum([J₀, opt.J_min])
        if J < J₀
            opt.J_min = J
            return Z + α * dZ
        end
        opt.J_min = J₀
    end

    while J > J₀ - β * α * d²
        if α < opt.cutoff
            println(io, "        α < cutoff")
            flush(io)
            return Z
        end

        α *= p

        y_final = experiment(
            A + α * dA,
            Δt + α * dΔt;
            only_final_measurement=true
        ).ys[end]

        J = opt.J(y_final)

        if opt.verbose
            println(io, "        J = $J, α = $α")
            flush(io)
        end
    end

    if opt.global_min
        opt.J_min = minimum([J, opt.J_min])
    end

    if manifold_projection
        if unitary
            n = Int(sqrt(Z.dims.Ũ⃗ ÷ 2))
            M = SpecialUnitary(n)
            Ũ⃗s = collect(eachcol(Z.Ũ⃗))
            dŨ⃗s = collect(eachcol(dZ.Ũ⃗))
            Us = iso_vec_to_operator.(Ũ⃗s)
            dUs = iso_vec_to_operator.(dŨ⃗s)
            Unews = [project(M, U + α * dU) for (U, dU) ∈ zip(Us, dUs)]
            Ũ⃗news = operator_to_iso_vec.(Unews)
            Ũ⃗new = hcat(Ũ⃗news...)
            Z_updated = Z + α * dZ
            update!(Z_updated, :Ũ⃗, Ũ⃗new)
            return Z_updated
        end
    else
        return Z + α * dZ
    end
end


end
