module PlottingFunctions

export save_animation
export plot_cosine_similarity

using ..Experiments
using ..ILCProblems

using NamedTrajectories

using CairoMakie
using LinearAlgebra

function save_animation(
    prob::ILCProblem,
    path::String;
    fps=2,
    plot_transformed_controls=false,
    markersize=10,
    plot_states=true,
)
    fig = Figure(resolution=(1000, 800))

    color = :glasbey_bw_minc_20_n256

    iter = Observable(0)

    ax_count = 0

    loss_ax = Axis(fig[1, :];
        title="Loss",
        xlabel="iter",
    )
    ax_count += 1

    ts = times(prob.Zs[1])

    lines!(loss_ax, prob.objective_history; color=:blue)

    Y_goal = hcat(prob.Y_goal.ys...)
    τs = prob.Y_goal.τs

    Y¹ = hcat(prob.Ys[1].ys...)

    Δy_max = maximum([maximum(abs.(hcat(Y.ys...) - Y_goal)) for Y ∈ prob.Ys])
    Δy_limit = 1.1 * abs(Δy_max)

    ΔY_ax = Axis(fig[ax_count + 1, :];
        title=@lift("|ΔY|: iter = $($iter)"),
        xlabel=L"t",
        limits=(nothing, (0.0, Δy_limit))
    )
    ax_count += 1

    ΔY¹ = abs.(Y¹ - Y_goal)

    ΔY_sp = series!(ΔY_ax, ts[τs], ΔY¹;
        color=color,
        markersize=markersize
    )

    if plot_states
        X_ax = Axis(fig[ax_count + 1, :];
            title=@lift("X: iter = $($iter)"),
            xlabel=L"t"
        )
        ax_count += 1

        X¹ = vcat([prob.Zs[1][name] for name ∈ prob.state_names]...);
        xdim = size(X¹, 1)

        X_sp = series!(X_ax, ts, X¹;
            color=color,
            labels=["x_$j" for j = 1:xdim]
        )
    end

    A_ax = Axis(fig[ax_count + 1, :];
        title=@lift("a(t): iter = $($iter)"),
        xlabel=L"t"
    )
    ax_count += 1

    max_bound = maximum(prob.Z_goal.bounds.a[2])
    ylims!(A_ax, [-max_bound, max_bound])

    A¹ = prob.Zs[1][prob.drive_name]

    A_sp = series!(A_ax, ts, A¹;
        labels=["a_$j" for j = 1:prob.Z_goal.dims[prob.drive_name]],
        color=color
    )

    if plot_transformed_controls
        A_trans_ax = Axis(fig[ax_count + 1, :]; title=@lift("a(t) quantized: iter = $($iter)"), xlabel=L"t")
        ax_count += 1

        A_trans_1 = prob.experiment.drive_transform(A¹)

        A_trans_sp = series!(A_trans_ax, ts, A_trans_1;
            labels=["a_$j" for j = 1:prob.Z_goal.dims.a]
        )
    end

    record(fig, path, 1:length(prob.Ys); framerate=fps) do i
        iter[] = i

        ts = times(prob.Zs[i])

        Yⁱ = hcat(prob.Ys[i].ys...)
        ΔYⁱ = abs.(Yⁱ - Y_goal)
        ΔY_sp[1] = ts[τs]
        ΔY_sp[2] = ΔYⁱ

        if plot_states
            Xⁱ = vcat([prob.Zs[i][name] for name ∈ prob.state_names]...);
            X_sp[1] = ts
            X_sp[2] = Xⁱ
            autolimits!(X_ax)
        end

        Aⁱ = prob.Zs[i][prob.drive_name]
        A_sp[1] = ts
        A_sp[2] = Aⁱ
        autolimits!(A_ax)


        if plot_transformed_controls
            A_trans_sp[2] = prob.experiment.drive_transform(Aⁱ)
        end
    end
end

save_animation(path::String, prob::ILCProblem; kwargs...) =
    save_animation(prob, path; kwargs...)


# plot cosine similarity for a vector of vectors
function plot_cosine_similarity(
    vecs::Vector{<:AbstractVector{<:Real}};
    title="Cosine similarity",
    xlabel="iter",
    ylabel="iter",
    kwargs...
)
    fig = Figure(resolution=(1000, 800))

    ax = Axis(fig[1, 1];
        title=title,
        xlabel=xlabel,
        ylabel=ylabel,
    )

    cos_sims = diagm(ones(length(vecs)))

    for (i, vec₁) in enumerate(vecs)
        for (j, vec₂) in enumerate(vecs)
            if i < j
                cos_sims[i, j] = dot(vec₁, vec₂) / (norm(vec₁) * norm(vec₂))
            end
        end
    end

    cos_sims = Symmetric(cos_sims)

    hm = heatmap!(ax, cos_sims)

    Colorbar(fig[:, end+1], hm)

    return fig
end

function plot_cosine_similarity(path::String, args...; kwargs...)
    fig = plot_cosine_similarity(args...; kwargs...)
    save(path, fig)
end


end
