module QuadraticProblems

export QuadraticProblem

export StaticQuadraticProblem
export DynamicQuadraticProblem

using ..Measurements
using ..Experiments

using NamedTrajectories
using TrajectoryIndexingUtils
using QuantumCollocation.Integrators
using QuantumCollocation.Dynamics
using LinearAlgebra
using SparseArrays
using ForwardDiff
using Einsum
using OSQP

abstract type QuadraticProblem end

struct DynamicQuadraticProblem <: QuadraticProblem
    ∂f::Function
    Cs::Vector{SparseMatrixCSC{Float64, Int64}} # jacobians of the measurement function g
    Hss::Union{Nothing,Vector{Vector{SparseMatrixCSC{Float64, Int64}}}} # hessians of g
    R::SparseMatrixCSC{Float64, Int64} # regularization hessian
    B::SparseMatrixCSC{Float64, Int64} # bounds constraint matrix
    quantum_state_names::Vector{Symbol} # names of quantum states
    KL::Bool
    verbose::Bool
    settings::Dict{Symbol, Any}
end

function DynamicQuadraticProblem(
    Z_goal::NamedTrajectory,
    quantum_state_names::Vector{Symbol},
    integrators::Vector{AbstractIntegrator},
    g::Union{Function, Vector{Function}},
    τs::Vector{Int},
    Q::Float64,
    Qf::Float64,
    R::Union{
        AbstractVector{Float64},
        NamedTuple{names, <:Tuple{Vararg{Float64}}} where names
    },
    objective_hessian::Bool,
    KL::Bool,
    verbose::Bool,
    io::IO,
    settings::Dict{Symbol, Any}
)
    @assert all([k ∈ keys(Z_goal.components) for k ∈ keys(R)]) "regularization multiplier keys must match trajectory component keys"

    ∂f = dynamics_jacobian(integrators, Z_goal)

    xf = vcat([Z_goal[end][name] for name ∈ quantum_state_names]...)
    xdim = length(xf)

    if g isa Function
        ∂g = x -> ForwardDiff.jacobian(g, x)
        ydim = length(g(xf))
    else
        # TODO: clean this up as it assumes states are adjacent and sequential
        g_stacked = x -> begin
            y = eltype(x)[]
            let x_stacked_comps = []
                let mark = 0
                    for name ∈quantum_state_names
                        push!(x_stacked_comps, (mark + 1):(mark + Z_goal.dims[name]))
                        mark += Z_goal.dims[name]
                    end
                end
                for (gᵢ, x_compsᵢ) ∈ zip(g, x_stacked_comps)
                    append!(y, gᵢ(x[x_compsᵢ]))
                end
            end
            return y
        end
        ∂g = x -> ForwardDiff.jacobian(g_stacked, x)
        ydim = length(g_stacked(xf))
    end

    if verbose
        println(io, "Building measurement jacobians...")
    end
    Cs = measurement_jacobians(∂g, Z_goal, quantum_state_names, τs, Q, Qf)
    if objective_hessian
        if verbose
            println(io, "Building measurement hessians...")
        end
        function ∂²g(x)
            H = ForwardDiff.jacobian(u -> vec(∂g(u)), x)
            return reshape(H, ydim, xdim, xdim)
        end
        Hss = measurement_hessians(∂²g, Z_goal, quantum_state_names, τs, Q, Qf)
    else
        Hss = nothing
    end

    if verbose
        println(io, "Building bounds constraint matrix...")
    end
    B_bounds = build_bounds_constraint_matrix(Z_goal)

    if verbose
        println(io, "Building initial/final value constraint matrix...")
    end
    B_init_final = build_initial_and_final_constraint_matrix(Z_goal)

    B = sparse_vcat(B_bounds, B_init_final)

    if verbose
        println(io, "Building regularization hessian...")
    end
    if R isa AbstractVector{Float64}
        @assert length(R) == Z_goal.dim * Z_goal.T "regularization vector must have length Z.dim * Z.T"
        R = diagm(R)
    else
        R = build_regularization_hessian(R, Z_goal)
    end

    return DynamicQuadraticProblem(
        ∂f,
        Cs,
        Hss,
        R,
        B,
        quantum_state_names,
        KL,
        verbose,
        settings
    )
end

function measurement_jacobians(
    ∂g::Function,
    Z::NamedTrajectory,
    quantum_state_names::Vector{Symbol},
    τs::Vector{Int},
    Q::Float64,
    Qf::Float64
)
    Cs = SparseMatrixCSC{Float64, Int}[]
    for τ ∈ τs
        x = vcat([Z[τ][name] for name ∈ quantum_state_names]...)
        if τ == Z.T
            C = sparse(∂g(x)) * Qf
        else
            C = sparse(∂g(x)) * Q
        end
        push!(Cs, C)
    end
    return Cs
end

function measurement_hessians(
    ∂²g::Function,
    Z::NamedTrajectory,
    quantum_state_names::Vector{Symbol},
    τs::Vector{Int},
    Q::Float64,
    Qf::Float64
)
    Hss = Vector{SparseMatrixCSC{Float64, Int}}[]
    for τ ∈ τs
        x = vcat([Z[τ][name] for name ∈ quantum_state_names]...)
        if τ == Z.T
            Hₜ = ∂²g(x) * Qf
        else
            Hₜ = ∂²g(x) * Q
        end
        Hₜs = [sparse(Hₜ[j, :, :]) for j ∈ axes(Hₜ, 1)]
        push!(Hss, Hₜs)
    end
    return Hss
end

function (QP::DynamicQuadraticProblem)(
    ΔY::DeterministicMeasurement,
    Ẑ::NamedTrajectory,
    io::IO=stdout
)
    model = OSQP.Model()

    # get the dynamics constraint jacobian
    if QP.verbose
        println(io, "Building dynamics constraint jacobian...")
    end
    ∂F = build_dynamics_constraint_jacobian(Ẑ, QP.∂f)

    # full constraint matrix
    A = sparse_vcat(∂F, QP.B)

    # get the constraint bounds
    if QP.verbose
        println(io, "Building constraint bounds...")
    end
    lb, ub = build_constraints(Ẑ)

    # get the hessian and gradient of the mle objective term
    if QP.verbose
        println(io, "Building hessian and gradient...")
    end

    ∇ = zeros(Ẑ.dim * Ẑ.T)

    x_comps = vcat([Ẑ.components[name] for name ∈ QP.quantum_state_names]...)

    for (τ, Cₜ, Δyₜ) ∈ zip(ΔY.τs, QP.Cs, ΔY.ys)
        ∇ₜ = Cₜ' * Δyₜ
        ∇[slice(τ, x_comps, Ẑ.dim)] = ∇ₜ
    end

    H = QP.R

    if !isnothing(QP.Hss)
        for (τ, Δyₜ, Cₜ, Hₜs) ∈ zip(ΔY.τs, ΔY.ys, QP.Cs, QP.Hss)
            Hₜ = Cₜ' * Cₜ
            for (Hₜᵢ, Δyₜᵢ) ∈ zip(Hₜs, Δyₜ)
                Hₜ += Δyₜᵢ * Hₜᵢ
            end
            H[slice(τ, x_comps, Ẑ.dim), slice(τ, x_comps, Ẑ.dim)] .+= Hₜ
        end
    end

    if QP.verbose
        println(io, "Setting up OSQP model...")
    end

    try
        OSQP.setup!(
            model;
            P=H,
            A=A,
            q=∇,
            l=lb,
            u=ub,
            QP.settings...
        )
    catch e
        throw(e)
    end

    if QP.verbose
        println(io, "Solving OSQP model...")
    end

    results = OSQP.solve!(model)

    if results.info.status != :Solved
        println(io, "WARNING: OSQP did not solve the problem, status: $(results.info.status)")
    end

    ΔZ = NamedTrajectory(results.x, Ẑ)

    return ΔZ
end


"""
    build_dynamics_constraint_jacobian(Ẑ::NamedTrajectory, ∂f::Function)

Builds the Jacobian of the dynamics constraint for the given trajectory
"""
@views function build_dynamics_constraint_jacobian(
    Z::NamedTrajectory,
    ∂f::Function
)
    ∂F = spzeros(
        Z.dims.states * (Z.T - 1),
        Z.dim * Z.T
    )
    for t = 1:Z.T - 1
        zₜ = Z[t].data
        zₜ₊₁ = Z[t+1].data
        ∂F[slice(t, Z.dims.states), slice(t:t+1, Z.dim)] = ∂f(zₜ, zₜ₊₁)
    end
    return sparse(∂F)
end



"""
    build_regularization_hessian(
        R::NamedTuple{names, Tuple{Vararg{Float64}}} where names,
        Z::NamedTrajectory
    )

Builds the hessian matrix for the regularization objective terms.
"""
@inline function build_regularization_hessian(
    R::NamedTuple{names, <:Tuple{Vararg{Float64}}} where names,
    Z::NamedTrajectory
)
    Hₜ = blockdiag([
        name ∈ keys(R) ?
            R[name] * sparse(I(Z.dims[name])) :
            spzeros(Z.dims[name], Z.dims[name])
        for name ∈ Z.names
    ]...)
    H = kron(sparse(I(Z.T)), Hₜ)
    return H
end

"""
    build_initial_and_final_constraint_matrix(Z::NamedTrajectory)

Builds and vcats the matrices to pick out the initial and final states of constrained components.
"""
@inline function build_initial_and_final_constraint_matrix(Z::NamedTrajectory)

    println("Building initial and final constraint matrices...")

    initCs = []

    for key ∈ keys(Z.initial)
        println("Building initial constraint matrix for $key...")
        C = spzeros(Z.dims[key], Z.dim * Z.T)
        C[:, slice(1, Z.components[key], Z.dim)] = sparse(1.0I(Z.dims[key]))
        push!(initCs, C)
    end

    finalCs = []

    for key ∈ keys(Z.final)
        println("Building final constraint matrix for $key...")
        C = spzeros(Z.dims[key], Z.dim * Z.T)
        C[:, slice(Z.T, Z.components[key], Z.dim)] = sparse(1.0I(Z.dims[key]))
        push!(finalCs, C)
    end

    Cs = [initCs; finalCs]

    if isempty(Cs)
        return spzeros(Float64, 0, Z.dim * Z.T)
    else
        return sparse_vcat(Cs...)
    end
end

"""
    build_initial_and_final_constraints(Z::NamedTrajectory)

Build the constraint vector for the initial and final states of constrained components.
"""
@inline function build_initial_and_final_constraints(Z::NamedTrajectory)
    if !isempty(keys(Z.initial))
        initial_cons = zeros(sum([Z.dims[key] for key ∈ keys(Z.initial)]))
    else
        initial_cons = zeros(0)
    end
    if !isempty(keys(Z.final))
        final_cons = zeros(sum([Z.dims[key] for key ∈ keys(Z.final)]))
    else
        final_cons = zeros(0)
    end
    return vcat(initial_cons, final_cons)
end


"""
    build_bounds_constraint_matrix(Z::NamedTrajectory)

Builds the bounds matrix by vcating the matrices that pick out the states/controls that are bounded in `Z`.

Acts on all timeslices: `t ∈ 1:T`
"""
@inline function build_bounds_constraint_matrix(Z::NamedTrajectory)
    Cs = []
    for key ∈ keys(Z.bounds)
        println("building constraint matrix for $key")
        if key == Z.timestep
            C = spzeros(Z.T, Z.dim * Z.T)
            for t = 1:Z.T
                C[t, index(t, Z.components[key][1], Z.dim)] = 1.0
            end
        else
            C = spzeros(Z.dims[key] * (Z.T - 2), Z.dim * Z.T)
            for (k, t) ∈ zip(1:Z.T-2, 2:Z.T-1)
                C[
                    slice(k, Z.dims[key]),
                    slice(t, Z.components[key], Z.dim)
                ] = sparse(I(Z.dims[key]))
            end
        end
        push!(Cs, C)
    end
    if isempty(Cs)
        return spzeros(Float64, 0, Z.dim * Z.T)
    else
        return sparse_vcat(Cs...)
    end
end

"""
    build_bounds_constraints(Z::NamedTrajectory)

Builds the bounds vector by vcating the bounds vectors for each bounded componenent in `Z`.

Acts on all timeslices: `t ∈ 1:T`
"""
@inline function build_bounds_constraints(Z::NamedTrajectory)
    lbs = []
    ubs = []
    for key ∈ keys(Z.bounds)
        if key == Z.timestep
            lb = vcat(fill(Z.bounds[key][1], Z.T)...)
            ub = vcat(fill(Z.bounds[key][2], Z.T)...)
        else
            lb = vcat(fill(Z.bounds[key][1], Z.T - 2)...) - vec(Z[key][:, 2:Z.T-1])
            ub = vcat(fill(Z.bounds[key][2], Z.T - 2)...) - vec(Z[key][:, 2:Z.T-1])
        end
        push!(lbs, lb)
        push!(ubs, ub)
    end
    if isempty(lbs)
        return zeros(0), zeros(0)
    else
        return vcat(lbs...), vcat(ubs...)
    end
end


"""
    build_constraints(Z::NamedTrajectory)

Builds the entire constraint vector -- i.e., dynamics, bounds, and initial/final constraints.
"""
@inline function build_constraints(Z::NamedTrajectory)
    ∂F_cons = zeros(Z.dims.states * (Z.T - 1))
    B_lb, B_ub = build_bounds_constraints(Z)
    C_init_final = build_initial_and_final_constraints(Z)
    lb = vcat(∂F_cons, B_lb, C_init_final)
    ub = vcat(∂F_cons, B_ub, C_init_final)
    return lb, ub
end

end
